import warnings
from abc import ABC
from typing import Any, Tuple, Optional, List

import math
import numpy as np
from numpy.typing import ArrayLike
from scipy import interpolate
import matplotlib.pyplot as plt

try:
    import cupy as cp

except (ModuleNotFoundError, ImportError):
    import numpy as cp

from . import detector as lisa_models
from .utils.utility import AET, get_array_module
from .utils.constants import *
from .stochastic import (
    StochasticContribution,
    FittedHyperbolicTangentGalacticForeground,
)
from .datacontainer import DataResidualArray
from .sensitivity import SensitivityMatrix
from .diagnostic import (
    noise_likelihood_term,
    residual_full_source_and_noise_likelihood,
    residual_source_likelihood_term,
    inner_product,
    data_signal_source_likelihood_term,
    data_signal_full_source_and_noise_likelihood,
)


class AnalysisContainer:
    """Combinatorial container that combines sensitivity and data information.

    Args:
        data_res_arr: Data / Residual / Signal array.
        sens_mat: Sensitivity information.
        signal_gen: Callable object that takes information through ``*args`` and ``**kwargs`` and
            generates a signal in the proper channel setup employed in ``data_res_arr`` and ``sens_mat``.

    """

    def __init__(
        self,
        data_res_arr: DataResidualArray,
        sens_mat: SensitivityMatrix,
        signal_gen: Optional[callable] = None,
    ) -> None:
        self.data_res_arr = data_res_arr
        self.sens_mat = sens_mat

        if signal_gen is not None:
            self.signal_gen = signal_gen

    @property
    def data_res_arr(self) -> DataResidualArray:
        """Data information."""
        return self._data_res_arr

    @data_res_arr.setter
    def data_res_arr(self, data_res_arr: DataResidualArray) -> None:
        """Set data."""
        assert isinstance(data_res_arr, DataResidualArray)
        self._data_res_arr = data_res_arr

    @property
    def sens_mat(self) -> SensitivityMatrix:
        """Sensitivity information."""
        return self._sens_mat

    @sens_mat.setter
    def sens_mat(self, sens_mat: SensitivityMatrix) -> None:
        "Set sensitivity information."
        assert isinstance(sens_mat, SensitivityMatrix)
        self._sens_mat = sens_mat

    @property
    def signal_gen(self) -> callable:
        """Signal generator."""
        if not hasattr(self, "_signal_gen"):
            raise ValueError(
                "User must input signal_gen kwarg to use the signal generator."
            )
        return self._signal_gen

    @signal_gen.setter
    def signal_gen(self, signal_gen: callable):
        """Set signal generator."""
        assert hasattr(signal_gen, "__call__")
        self._signal_gen = signal_gen

    def loglog(self) -> Tuple[plt.Figure, plt.Axes]:
        """Produce loglog plot of both source and sensitivity information.

        Returns:
            Matplotlib figure and axes object in a 2-tuple.

        """
        fig, ax = self.sens_mat.loglog(char_strain=True)
        if self.sens_mat.ndim == 3:
            # 3x3 most likely
            for i in range(self.sens_mat.shape[0]):
                for j in range(i, self.sens_mat.shape[1]):
                    # char strain
                    ax[i * self.sens_mat.shape[1] + j].loglog(
                        self.data_res_arr.f_arr,
                        self.data_res_arr.f_arr * self.data_res_arr[i],
                    )
                    ax[i * self.sens_mat.shape[1] + j].loglog(
                        self.data_res_arr.f_arr,
                        self.data_res_arr.f_arr * self.data_res_arr[j],
                    )
        else:
            for i in range(self.sens_mat.shape[0]):
                ax[i].loglog(
                    self.data_res_arr.f_arr,
                    self.data_res_arr.f_arr * self.data_res_arr[i],
                )
        return (fig, ax)

    def inner_product(self, **kwargs: dict) -> float | complex:
        """Return the inner product of the current set of information

        Args:
            **kwargs: Inner product keyword arguments.

        Returns:
            Inner product value.

        """
        if "psd" in kwargs:
            kwargs.pop("psd")

        return inner_product(self.data_res_arr, self.data_res_arr, psd=self.sens_mat)

    def snr(self, **kwargs: dict) -> float:
        """Return the SNR of the current set of information

        Args:
            **kwargs: Inner product keyword arguments.

        Returns:
            SNR value.

        """
        return self.inner_product(**kwargs).real ** (1 / 2)

    def template_inner_product(
        self, template: DataResidualArray, **kwargs: dict
    ) -> float | complex:
        """Calculate the inner product of a template with the data.

        Args:
            template: Template signal.
            **kwargs: Keyword arguments to pass to :func:`lisatools.diagnostic.inner_product`.

        Returns:
            Inner product value.

        """
        if "psd" in kwargs:
            kwargs.pop("psd")

        if "include_psd_info" in kwargs:
            kwargs.pop("include_psd_info")

        ip_val = inner_product(self.data_res_arr, template, psd=self.sens_mat, **kwargs)
        return ip_val

    def template_snr(
        self, template: DataResidualArray, phase_maximize: bool = False, **kwargs: dict
    ) -> Tuple[float, float]:
        """Calculate the SNR of a template, both optimal and detected.

        Args:
            template: Template signal.
            phase_maximize: If ``True``, maximize over an overall phase.
            **kwargs: Keyword arguments to pass to :func:`lisatools.diagnostic.inner_product`.

        Returns:
            ``(optimal snr, detected snr)``.

        """
        kwargs_in = kwargs.copy()
        if "psd" in kwargs:
            kwargs_in.pop("psd")

        if "complex" in kwargs_in:
            kwargs_in.pop("complex")

        # TODO: should we cache?
        h_h = inner_product(template, template, psd=self.sens_mat, **kwargs_in)
        non_marg_d_h = inner_product(
            self.data_res_arr, template, psd=self.sens_mat, complex=True, **kwargs_in
        )
        d_h = np.abs(non_marg_d_h) if phase_maximize else non_marg_d_h.copy()
        self.non_marg_d_h = non_marg_d_h

        opt_snr = np.sqrt(h_h.real)
        det_snr = d_h.real / opt_snr
        return (opt_snr, det_snr)

    def template_likelihood(
        self,
        template: DataResidualArray,
        include_psd_info: bool = False,
        phase_maximize: bool = False,
        **kwargs: dict
    ) -> float:
        """Calculate the Likelihood of a template against the data.

        Args:
            template: Template signal.
            include_psd_info: If ``True``, add the PSD term to the Likelihood value.
            phase_maximize: If ``True``, maximize over an overall phase.
            **kwargs: Keyword arguments to pass to :func:`lisatools.diagnostic.inner_product`.

        Returns:
            Likelihood value.

        """
        kwargs_in = kwargs.copy()
        if "psd" in kwargs_in:
            kwargs_in.pop("psd")

        if "complex" in kwargs_in:
            kwargs_in.pop("complex")

        # TODO: should we cache?
        d_d = inner_product(
            self.data_res_arr, self.data_res_arr, psd=self.sens_mat, **kwargs_in
        )
        h_h = inner_product(template, template, psd=self.sens_mat, **kwargs_in)
        non_marg_d_h = inner_product(
            self.data_res_arr, template, psd=self.sens_mat, complex=True, **kwargs_in
        )
        d_h = np.abs(non_marg_d_h) if phase_maximize else non_marg_d_h.copy()
        self.non_marg_d_h = non_marg_d_h
        like_out = -1 / 2 * (d_d + h_h - 2 * d_h).real

        if include_psd_info:
            # add noise term if requested
            like_out += self.likelihood(noise_only=True)

        return like_out

    def likelihood(
        self, source_only: bool = False, noise_only: bool = False, **kwargs: dict
    ) -> float | complex:
        """Return the likelihood of the current arangement.

        Args:
            source_only: If ``True`` return the source-only Likelihood.
            noise_only: If ``True``, return the noise part of the Likelihood alone.
            **kwargs: Keyword arguments to pass to :func:`lisatools.diagnostic.inner_product`.

        Returns:
            Likelihood value.

        """
        if noise_only and source_only:
            raise ValueError("noise_only and source only cannot both be True.")
        elif noise_only:
            return noise_likelihood_term(self.sens_mat)
        elif source_only:
            return residual_source_likelihood_term(
                self.data_res_arr, psd=self.sens_mat, **kwargs
            )
        else:
            return residual_full_source_and_noise_likelihood(
                self.data_res_arr, self.sens_mat, **kwargs
            )

    def _calculate_signal_operation(
        self,
        calc: str,
        *args: ArrayLike,
        source_only: bool = False,
        waveform_kwargs: Optional[dict] = {},
        data_res_arr_kwargs: Optional[dict] = {},
        **kwargs: dict
    ) -> float | complex:
        """Return the likelihood of a generated signal with the data.

        Args:
            calc: Type of calculation to do. Options are ``"likelihood"``, ``"inner_product"``, or ``"snr"``.
            *args: Arguments to waveform generating function. Must include parameters.
            source_only: If ``True`` return the source-only Likelihood (leave out noise part).
            waveform_kwargs: Keyword arguments to pass to waveform generator.
            data_res_arr_kwargs: Keyword arguments for instantiation of :class:`DataResidualArray`.
                This can be used if any transforms are desired prior to the Likelihood computation. If it is not input,
                the kwargs are taken to be the same as those used to initalize ``self.data_res_arr``.
            **kwargs: Keyword arguments to pass to :func:`lisatools.diagnostic.inner_product`

        Returns:
            Likelihood value.

        """

        if data_res_arr_kwargs == {}:
            data_res_arr_kwargs = self.data_res_arr.init_kwargs

        template = DataResidualArray(
            self.signal_gen(*args, **waveform_kwargs), **data_res_arr_kwargs
        )

        args_2 = (template,)

        if "include_psd_info" in kwargs:
            assert kwargs["include_psd_info"] == (not source_only)
            kwargs.pop("include_psd_info")

        kwargs = dict(psd=self.sens_mat, **kwargs)

        if calc == "likelihood":
            kwargs["include_psd_info"] = not source_only
            return self.template_likelihood(*args_2, **kwargs)
        elif calc == "inner_product":
            return self.template_inner_product(*args_2, **kwargs)
        elif calc == "snr":
            return self.template_snr(*args_2, **kwargs)
        else:
            raise ValueError("`calc` must be 'likelihood', 'inner_product', or 'snr'.")

    def calculate_signal_likelihood(
        self,
        *args: ArrayLike,
        source_only: bool = False,
        waveform_kwargs: Optional[dict] = {},
        data_res_arr_kwargs: Optional[dict] = {},
        **kwargs: dict
    ) -> float | complex:
        """Return the likelihood of a generated signal with the data.

        Args:
            params: Arguments to waveform generating function. Must include parameters.
            source_only: If ``True`` return the source-only Likelihood (leave out noise part).
            waveform_kwargs: Keyword arguments to pass to waveform generator.
            data_res_arr_kwargs: Keyword arguments for instantiation of :class:`DataResidualArray`.
                This can be used if any transforms are desired prior to the Likelihood computation.
            **kwargs: Keyword arguments to pass to :func:`lisatools.diagnostic.inner_product`

        Returns:
            Likelihood value.

        """

        return self._calculate_signal_operation(
            "likelihood",
            *args,
            source_only=source_only,
            waveform_kwargs=waveform_kwargs,
            data_res_arr_kwargs=data_res_arr_kwargs,
            **kwargs
        )

    def calculate_signal_inner_product(
        self,
        *args: ArrayLike,
        source_only: bool = False,
        waveform_kwargs: Optional[dict] = {},
        data_res_arr_kwargs: Optional[dict] = {},
        **kwargs: dict
    ) -> float | complex:
        """Return the inner product of a generated signal with the data.

        Args:
            *args: Arguments to waveform generating function. Must include parameters.
            source_only: If ``True`` return the source-only Likelihood (leave out noise part).
            waveform_kwargs: Keyword arguments to pass to waveform generator.
            data_res_arr_kwargs: Keyword arguments for instantiation of :class:`DataResidualArray`.
                This can be used if any transforms are desired prior to the Likelihood computation.
            **kwargs: Keyword arguments to pass to :func:`lisatools.diagnostic.inner_product`

        Returns:
            Inner product value.

        """

        return self._calculate_signal_operation(
            "inner_product",
            *args,
            source_only=source_only,
            waveform_kwargs=waveform_kwargs,
            data_res_arr_kwargs=data_res_arr_kwargs,
            **kwargs
        )

    def calculate_signal_snr(
        self,
        *args: ArrayLike,
        source_only: bool = False,
        waveform_kwargs: Optional[dict] = {},
        data_res_arr_kwargs: Optional[dict] = {},
        **kwargs: dict
    ) -> Tuple[float, float]:
        """Return the SNR of a generated signal with the data.

        Args:
            *args: Arguments to waveform generating function. Must include parameters.
            source_only: If ``True`` return the source-only Likelihood (leave out noise part).
            waveform_kwargs: Keyword arguments to pass to waveform generator.
            data_res_arr_kwargs: Keyword arguments for instantiation of :class:`DataResidualArray`.
                This can be used if any transforms are desired prior to the Likelihood computation.
            **kwargs: Keyword arguments to pass to :func:`lisatools.diagnostic.inner_product`

        Returns:
            Snr values (optimal, detected).

        """

        return self._calculate_signal_operation(
            "snr",
            *args,
            source_only=source_only,
            waveform_kwargs=waveform_kwargs,
            data_res_arr_kwargs=data_res_arr_kwargs,
            **kwargs
        )

    def eryn_likelihood_function(self, x, *args, **kwargs):
        if x.ndim == 1:
            input_vals = tuple(x) + tuple(args)
            return self.calculate_signal_likelihood(*input_vals, **kwargs)
        elif x.ndim == 2:
            likelihood_out = np.zeros(x.shape[0])
            for i in range(x.shape[0]):
                input_vals = tuple(x[i]) + tuple(args)
                likelihood_out[i] = self.calculate_signal_likelihood(
                    *input_vals, **kwargs
                )

        else:
            raise ValueError("x must be a 1D or 2D array.")



# -*- coding: utf-8 -*-

from copy import deepcopy
from inspect import Attribute
import numpy as np
from scipy import stats
import warnings
import time
from gbgpu.utils.utility import get_N
from lisatools.sensitivity import Soms_d_all, Sa_a_all

try:
    import cupy as xp

    gpu_available = True
except ModuleNotFoundError:
    import numpy as xp

    gpu_available = False

from lisatools.utils.utility import searchsorted2d_vec, get_groups_from_band_structure
from eryn.moves import StretchMove
from eryn.prior import ProbDistContainer
from eryn.utils.utility import groups_from_inds
from eryn.utils import PeriodicContainer

from eryn.moves import GroupStretchMove, Move
from eryn.moves.multipletry import logsumexp, get_mt_computations

from ...diagnostic import inner_product
from lisatools.globalfit.state import State
from lisatools.sampling.prior import GBPriorWrap


__all__ = ["GBSpecialStretchMove"]


# MHMove needs to be to the left here to overwrite GBBruteRejectionRJ RJ proposal method
class GBSpecialStretchMove(GroupStretchMove, Move):
    """Generate Revesible-Jump proposals for GBs with try-force rejection

    Will use gpu if template generator uses GPU.

    Args:
        priors (object): :class:`ProbDistContainer` object that has ``logpdf``
            and ``rvs`` methods.

    """

    def __init__(
        self,
        gb,
        priors,
        start_freq_ind,
        data_length,
        mgh,
        fd,
        band_edges,
        gpu_priors,
        *args,
        waveform_kwargs={},
        parameter_transforms=None,
        snr_lim=1e-10,
        rj_proposal_distribution=None,
        num_repeat_proposals=1,
        name=None,
        use_prior_removal=False,
        phase_maximize=False, 
        **kwargs
    ):
        # return_gpu is a kwarg for the stretch move
        GroupStretchMove.__init__(self, *args, return_gpu=True, **kwargs)

        self.gpu_priors = gpu_priors
        self.name = name
        self.num_repeat_proposals = num_repeat_proposals
        self.use_prior_removal = use_prior_removal

        for key in priors:
            if not isinstance(priors[key], ProbDistContainer) and not isinstance(priors[key], GBPriorWrap):
                raise ValueError(
                    "Priors need to be eryn.priors.ProbDistContainer object."
                )
        
        self.priors = priors
        self.gb = gb
        self.stop_here = True

        args = [priors["gb"].priors_in[(0, 1)].rho_star]
        args += [priors["gb"].priors_in[(0, 1)].frequency_prior.min_val, priors["gb"].priors_in[(0, 1)].frequency_prior.max_val]
        for i in range(2, 8):
            args += [priors["gb"].priors_in[i].min_val, priors["gb"].priors_in[i].max_val]
        
        self.gpu_cuda_priors = self.gb.pyPriorPackage(*tuple(args))
        self.gpu_cuda_wrap = self.gb.pyPeriodicPackage(2 * np.pi, np.pi, 2 * np.pi)

        # use gpu from template generator
        # self.use_gpu = gb.use_gpu
        if self.use_gpu:
            self.mempool = self.xp.get_default_memory_pool()

        self.band_edges = band_edges
        self.num_bands = len(band_edges) - 1
        self.start_freq_ind = start_freq_ind
        self.data_length = data_length
        self.waveform_kwargs = waveform_kwargs
        self.parameter_transforms = parameter_transforms
        self.fd = fd
        self.df = (fd[1] - fd[0]).item()
        self.mgh = mgh
        self.phase_maximize = phase_maximize

        self.snr_lim = snr_lim

        self.band_edges = self.xp.asarray(self.band_edges)

        self.rj_proposal_distribution = rj_proposal_distribution
        self.is_rj_prop = self.rj_proposal_distribution is not None

        # setup N vals for bands
        band_mean_f = (self.band_edges[1:] + self.band_edges[:-1]).get() / 2
        self.band_N_vals = xp.asarray(get_N(np.full_like(band_mean_f, 1e-30), band_mean_f, self.waveform_kwargs["T"], self.waveform_kwargs["oversample"]))

    def setup_gbs(self, branch):
        st = time.perf_counter()
        coords = branch.coords
        inds = branch.inds
        supps = branch.branch_supplimental
        ntemps, nwalkers, nleaves_max, ndim = branch.shape
        all_remaining_freqs = coords[0][inds[0]][:, 1]

        all_remaining_cords = coords[0][inds[0]]

        num_remaining = len(all_remaining_freqs)

        all_temp_fs = self.xp.asarray(coords[inds][:, 1])

        # TODO: improve this?
        self.inds_freqs_sorted = self.xp.asarray(np.argsort(all_remaining_freqs))
        self.freqs_sorted = self.xp.asarray(np.sort(all_remaining_freqs))
        self.all_coords_sorted = self.xp.asarray(all_remaining_cords)[
            self.inds_freqs_sorted
        ]

        total_binaries = inds.sum().item()
        still_going = xp.ones(total_binaries, dtype=bool) 
        inds_zero = xp.searchsorted(self.freqs_sorted, all_temp_fs, side="right") - 1
        left_inds = inds_zero - int(self.nfriends / 2)
        right_inds = inds_zero + int(self.nfriends / 2) - 1

        # do right first here
        right_inds[left_inds < 0] = self.nfriends - 1
        left_inds[left_inds < 0] = 0
        
        # do left first here
        left_inds[right_inds > len(self.freqs_sorted) - 1] = len(self.freqs_sorted) - self.nfriends
        right_inds[right_inds > len(self.freqs_sorted) - 1] = len(self.freqs_sorted) - 1

        assert np.all(right_inds - left_inds == self.nfriends - 1)
        assert not np.any(right_inds < 0) and not np.any(right_inds > len(self.freqs_sorted) - 1) and not np.any(left_inds < 0) and not np.any(left_inds > len(self.freqs_sorted) - 1)
        
        jjj = 0
        while np.any(still_going):
            distance_left = np.abs(all_temp_fs[still_going] - self.freqs_sorted[left_inds[still_going]])
            distance_right = np.abs(all_temp_fs[still_going] - self.freqs_sorted[right_inds[still_going]])

            check_move_right = (distance_right <= distance_left)
            check_left_inds = left_inds[still_going][check_move_right] + 1
            check_right_inds = right_inds[still_going][check_move_right] + 1

            new_distance_right = np.abs(all_temp_fs[still_going][check_move_right] - self.freqs_sorted[check_right_inds])

            change_inds = xp.arange(len(all_temp_fs))[still_going][check_move_right][(new_distance_right < distance_left[check_move_right]) & (check_right_inds < len(self.freqs_sorted))]

            left_inds[change_inds] += 1
            right_inds[change_inds] += 1

            stop_inds_right_1 = xp.arange(len(all_temp_fs))[still_going][check_move_right][(check_right_inds >= len(self.freqs_sorted))]

            # last part is just for up here, below it will remove if it is still equal
            stop_inds_right_2 = xp.arange(len(all_temp_fs))[still_going][check_move_right][(new_distance_right >= distance_left[check_move_right]) & (check_right_inds < len(self.freqs_sorted)) & (distance_right[check_move_right] != distance_left[check_move_right])]
            stop_inds_right = xp.concatenate([stop_inds_right_1, stop_inds_right_2])
            assert np.all(still_going[stop_inds_right])

            # equal to should only be left over if it was equal above and moving right did not help
            check_move_left = (distance_left <= distance_right)
            check_left_inds = left_inds[still_going][check_move_left] - 1
            check_right_inds = right_inds[still_going][check_move_left] - 1

            new_distance_left = np.abs(all_temp_fs[still_going][check_move_left] - self.freqs_sorted[check_left_inds])
            
            change_inds = xp.arange(len(all_temp_fs))[still_going][check_move_left][(new_distance_left < distance_right[check_move_left]) & (check_left_inds >= 0)]

            left_inds[change_inds] -= 1
            right_inds[change_inds] -= 1

            stop_inds_left_1 = xp.arange(len(all_temp_fs))[still_going][check_move_left][(check_left_inds < 0)]
            stop_inds_left_2 = xp.arange(len(all_temp_fs))[still_going][check_move_left][(new_distance_left >= distance_right[check_move_left]) & (check_left_inds >= 0)]
            stop_inds_left = xp.concatenate([stop_inds_left_1, stop_inds_left_2])
            
            stop_inds = xp.concatenate([stop_inds_right, stop_inds_left])
            still_going[stop_inds] = False
            # print(jjj, still_going.sum())
            if jjj >= self.nfriends:
                breakpoint()
            jjj += 1

        start_inds = left_inds.copy().get()

        start_inds_all = np.zeros_like(inds, dtype=np.int32)
        start_inds_all[inds] = start_inds.astype(np.int32)

        if "friend_start_inds" not in supps:
            supps.add_objects({"friend_start_inds": start_inds_all})
        else:
            supps[:] = {"friend_start_inds": start_inds_all}

        self.stretch_friends_args_in = tuple([tmp.copy() for tmp in self.all_coords_sorted.T])
        et = time.perf_counter()
        self.mempool.free_all_blocks()
        # print("SETUP:", et - st)
        # start_inds_freq_out = np.zeros((ntemps, nwalkers, nleaves_max), dtype=int)
        # freqs_sorted_here = self.freqs_sorted.get()
        # freqs_remaining_here = all_remaining_freqs

        # start_ind_best = np.zeros_like(freqs_remaining_here, dtype=int)

        # best_index = (
        #     np.searchsorted(freqs_sorted_here, freqs_remaining_here, side="right") - 1
        # )
        # best_index[best_index < self.nfriends] = self.nfriends
        # best_index[best_index >= len(freqs_sorted_here) - self.nfriends] = (
        #     len(freqs_sorted_here) - self.nfriends
        # )
        # check_inds = (
        #     best_index[:, None]
        #     + np.tile(np.arange(2 * self.nfriends), (best_index.shape[0], 1))
        #     - self.nfriends
        # )

        # check_freqs = freqs_sorted_here[check_inds]
        # breakpoint()

        # # batch_count = 1000
        # # split_inds = np.arange(batch_count, freqs_remaining_here.shape[0], batch_count)

        # # splits_remain = np.split(freqs_remaining_here, split_inds)
        # # splits_check = np.split(check_freqs, split_inds)

        # # out = []
        # # for i, (split_r, split_c) in enumerate(zip(splits_remain, splits_check)):
        # #     out.append(np.abs(split_r[:, None] - split_c))
        # #     print(i)

        # # freq_distance = np.asarray(out)

        # freq_distance = np.abs(freqs_remaining_here[:, None] - check_freqs)
        # breakpoint()

        # keep_min_inds = np.argsort(freq_distance, axis=-1)[:, : self.nfriends].min(
        #     axis=-1
        # )
        # start_inds_freq = check_inds[(np.arange(len(check_inds)), keep_min_inds)]

        # start_inds_freq_out[inds] = start_inds_freq

        # start_inds_freq_out[~inds] = -1

        # if "friend_start_inds" not in supps:
        #     supps.add_objects({"friend_start_inds": start_inds_freq_out})
        # else:
        #     supps[:] = {"friend_start_inds": start_inds_freq_out}

        # self.all_friends_start_inds_sorted = self.xp.asarray(
        #     start_inds_freq_out[inds][self.inds_freqs_sorted.get()]
        # )

    def find_friends(self, name, gb_points_to_move, s_inds=None):
        if s_inds is None:
            raise ValueError

        inds_points_to_move = self.xp.asarray(s_inds.flatten())

        half_friends = int(self.nfriends / 2)

        gb_points_for_move = gb_points_to_move.reshape(-1, 8).copy()

        if not hasattr(self, "ntemps"):
            self.ntemps = 1

        inds_start_freq_to_move = self.current_friends_start_inds[
            inds_points_to_move.reshape(self.ntemps, self.nwalkers, -1)
        ]

        deviation = self.xp.random.randint(
            0, self.nfriends, size=len(inds_start_freq_to_move)
        )

        inds_keep_friends = inds_start_freq_to_move + deviation

        inds_keep_friends[inds_keep_friends < 0] = 0
        inds_keep_friends[inds_keep_friends >= len(self.all_coords_sorted)] = (
            len(self.all_coords_sorted) - 1
        )

        gb_points_for_move[inds_points_to_move] = self.all_coords_sorted[
            inds_keep_friends
        ]

        return gb_points_for_move.reshape(self.ntemps, -1, 1, 8)

    def new_find_friends(self, name, inds_in):
        inds_start_freq_to_move = self.current_friends_start_inds[tuple(inds_in)]

        deviation = self.xp.random.randint(
            0, self.nfriends, size=len(inds_start_freq_to_move)
        )

        inds_keep_friends = inds_start_freq_to_move + deviation

        inds_keep_friends[inds_keep_friends < 0] = 0
        inds_keep_friends[inds_keep_friends >= len(self.all_coords_sorted)] = (
            len(self.all_coords_sorted) - 1
        )

        gb_points_for_move = self.all_coords_sorted[
            inds_keep_friends
        ]

        return gb_points_for_move

    def setup(self, branches):
        for i, (name, branch) in enumerate(branches.items()):
            if name != "gb":
                continue

            if not self.is_rj_prop and self.time % self.n_iter_update == 0:
                self.setup_gbs(branch)

            elif self.is_rj_prop and self.time == 0:
                ndim = branch.shape[-1]
                self.stretch_friends_args_in = tuple([xp.array([]) for _ in range(ndim)])

            # update any shifted start inds due to tempering (need to do this every non-rj move)
            """if not self.is_rj_prop:
                # fix the ones that have been added in RJ
                fix = (
                    branch.branch_supplimental.holder["friend_start_inds"][:] == -1
                ) & branch.inds

                if np.any(fix):
                    new_freqs = xp.asarray(branch.coords[fix][:, 1])
                    # TODO: is there a better way of doing this?

                    # fill information into friend finder for new binaries
                    branch.branch_supplimental.holder["friend_start_inds"][fix] = (
                        (
                            xp.searchsorted(self.freqs_sorted, new_freqs, side="right")
                            - 1
                        )
                        * (
                            (new_freqs > self.freqs_sorted[0])
                            & (new_freqs < self.freqs_sorted[-1])
                        )
                        + 0 * (new_freqs < self.freqs_sorted[0])
                        + (len(self.freqs_sorted) - 1)
                        * (new_freqs > self.freqs_sorted[-1])
                    ).get()

                # make sure current start inds reflect alive binaries
                self.current_friends_start_inds = self.xp.asarray(
                    branch.branch_supplimental.holder["friend_start_inds"][:]
                )
            """

            self.mempool.free_all_blocks()

    def propose(self, model, state):
        """Use the move to generate a proposal and compute the acceptance

        Args:
            model (:class:`eryn.model.Model`): Carrier of sampler information.
            state (:class:`State`): Current state of the sampler.

        Returns:
            :class:`State`: State of sampler after proposal is complete.

        """
        st = time.perf_counter()

        self.xp.cuda.runtime.setDevice(self.mgh.gpus[0])

        self.current_state = state
        np.random.seed(10)
        # print("start stretch")

        # Check that the dimensions are compatible.
        ndim_total = 0
        for branch in state.branches.values():
            ntemps, nwalkers, nleaves_, ndim_ = branch.shape
            ndim_total += ndim_ * nleaves_

        self.nwalkers = nwalkers
        
        # Run any move-specific setup.
        self.setup(state.branches)

        new_state = State(state, copy=True)
        band_temps = xp.asarray(state.band_info["band_temps"].copy())

        if self.is_rj_prop:
            orig_store = new_state.log_like[0].copy()
        gb_coords = xp.asarray(new_state.branches["gb"].coords)

        self.mempool.free_all_blocks()

        # self.mgh.map = new_state.supplimental.holder["overall_inds"].flatten()

        # data should not be whitened

        ntemps, nwalkers, nleaves_max, ndim = state.branches_coords["gb"].shape

        group_temp_finder = [
            self.xp.repeat(self.xp.arange(ntemps), nwalkers * nleaves_max).reshape(
                ntemps, nwalkers, nleaves_max
            ),
            self.xp.tile(self.xp.arange(nwalkers), (ntemps, nleaves_max, 1)).transpose(
                (0, 2, 1)
            ),
            self.xp.tile(self.xp.arange(nleaves_max), ((ntemps, nwalkers, 1))),
        ]

        """(
            gb_coords,
            gb_inds_orig,
            points_curr,
            prior_all_curr,
            gb_inds,
            N_vals_in,
            prop_to_curr_map,
            factors,
            inds_curr,
            proposal_specific_information
        ) = self.get_special_proposal_setup(model, new_state, state, group_temp_finder)"""

        waveform_kwargs_now = self.waveform_kwargs.copy()
        if "N" in waveform_kwargs_now:
            waveform_kwargs_now.pop("N")
        waveform_kwargs_now["start_freq_ind"] = self.start_freq_ind

        # if self.is_rj_prop:
        #     print("START:", new_state.log_like[0])
        log_like_tmp = self.xp.asarray(new_state.log_like)
        log_prior_tmp = self.xp.asarray(new_state.log_prior)

        self.mempool.free_all_blocks()

        gb_inds = self.xp.asarray(new_state.branches["gb"].inds)
        gb_inds_orig = gb_inds.copy()

        data = self.mgh.data_list
        base_data = self.mgh.base_data_list
        psd = self.mgh.psd_list
        lisasens = self.mgh.lisasens_list

        # do unique for band size as separator between asynchronous kernel launches
        # band_indices = self.xp.asarray(new_state.branches["gb"].branch_supplimental.holder["band_inds"])
        band_indices = self.xp.searchsorted(self.band_edges, xp.asarray(new_state.branches["gb"].coords[:, :, :, 1]).flatten() / 1e3, side="right").reshape(new_state.branches["gb"].coords[:, :, :, 1].shape) - 1
            
        # N_vals_in = self.xp.asarray(new_state.branches["gb"].branch_supplimental.holder["N_vals"])
        points_curr = self.xp.asarray(new_state.branches["gb"].coords)
        points_curr_orig = points_curr.copy()
        # N_vals_in_orig = N_vals_in.copy()
        band_indices_orig = band_indices.copy()
        
        if self.is_rj_prop:
            if isinstance(self.rj_proposal_distribution["gb"], list):
                raise NotImplementedError
                assert len(self.rj_proposal_distribution["gb"]) == ntemps
                proposal_logpdf = xp.zeros((points_curr.shape[0], np.prod(points_curr.shape[1:-1])))
                for t in range(ntemps):
                    new_sources = xp.full_like(points_curr[t, ~gb_inds[t]], np.nan)
                    fix = xp.full(new_sources.shape[0], True)
                    while xp.any(fix):
                        new_sources[fix] = self.rj_proposal_distribution["gb"][t].rvs(size=fix.sum().item())
                        fix = xp.any(xp.isnan(new_sources), axis=-1)
                    points_curr[t, ~gb_inds[t]] = new_sources
                    band_indices[t, ~gb_inds[t]] = self.xp.searchsorted(self.band_edges, new_sources[:, 1] / 1e3, side="right") - 1
                    # change gb_inds
                    gb_inds[t, :] = True
                    assert np.all(gb_inds[t])
                    proposal_logpdf[t] = self.rj_proposal_distribution["gb"][t].logpdf(
                        points_curr[t, gb_inds[t]]
                    )

                proposal_logpdf = proposal_logpdf.flatten().copy()
            
            else:
                new_sources = xp.full_like(points_curr[~gb_inds], np.nan)
                fix = xp.full(new_sources.shape[0], True)
                while xp.any(fix):
                    if self.name == "rj_prior":
                        new_sources[fix] = self.rj_proposal_distribution["gb"].rvs(size=fix.sum().item(), psds=self.mgh.psd_shaped[0][0], walker_inds=group_temp_finder[1][~gb_inds][fix])
                    else:
                        new_sources[fix] = self.rj_proposal_distribution["gb"].rvs(size=fix.sum().item())
                    
                    fix = xp.any(xp.isnan(new_sources), axis=-1)
                points_curr[~gb_inds] = new_sources

                band_indices[~gb_inds] = self.xp.searchsorted(self.band_edges, new_sources[:, 1] / 1e3, side="right") - 1
                # N_vals_in[~gb_inds] = self.xp.asarray(
                #     get_N(
                #         xp.full_like(new_sources[:, 1], 1e-30),
                #         new_sources[:, 1] / 1e3,
                #         self.waveform_kwargs["T"],
                #         self.waveform_kwargs["oversample"],
                #         xp=self.xp
                #     )
                # )

                # change gb_inds
                gb_inds[:] = True
                assert np.all(gb_inds)
                if self.name == "rj_prior":
                    proposal_logpdf = self.rj_proposal_distribution["gb"].logpdf(
                        points_curr[gb_inds], psds=self.mgh.psd_shaped[0][0], walker_inds=group_temp_finder[1][gb_inds]
                    )
                else:
                    proposal_logpdf = self.rj_proposal_distribution["gb"].logpdf(
                        points_curr[gb_inds]
                    )

            factors = (proposal_logpdf * -1) * (~gb_inds_orig).flatten() + (proposal_logpdf * +1) * (gb_inds_orig).flatten()

            if self.name == "rj_prior" and self.use_prior_removal:
                factors[~gb_inds_orig.flatten()] = -1e300
        else:
            factors = xp.zeros(gb_inds_orig.sum().item())

        start_inds_all = xp.asarray(new_state.branches["gb"].branch_supplimental.holder["friend_start_inds"], dtype=xp.int32)[gb_inds]
        points_curr = points_curr[gb_inds]
        # N_vals_in = N_vals_in[gb_inds]
        band_indices = band_indices[gb_inds]
        gb_inds_in = gb_inds_orig[gb_inds]

        temp_indices = group_temp_finder[0][gb_inds]
        walker_indices = group_temp_finder[1][gb_inds]
        leaf_indices = group_temp_finder[2][gb_inds]
        
        unique_N = self.xp.unique(self.band_N_vals)
        # remove 0
        unique_N = unique_N[unique_N != 0]

        do_synchronize = False
        device = self.xp.cuda.runtime.getDevice()

        units = 2 if not self.is_rj_prop else 2
        # random start to rotation around 
        start_unit = model.random.randint(units)

        ll_after = (
            self.mgh.get_ll(include_psd_info=True)
        )
        # print(np.abs(new_state.log_like - ll_after).max())
        store_max_diff = np.abs(new_state.log_like[0] - ll_after).max()
        # print("CHECKING 0:", store_max_diff, self.is_rj_prop)
        self.check_ll_inject(new_state)

        per_walker_band_proposals = xp.zeros((ntemps, nwalkers, self.num_bands), dtype=int)
        per_walker_band_accepted = xp.zeros((ntemps, nwalkers, self.num_bands), dtype=int)
        
        total_keep = 0
        for tmp in range(units):
            remainder = (start_unit + tmp) % units
            all_inputs = []
            band_bookkeep_info = []
            indiv_info = []
            current_parameter_arrays = []
            N_vals_list = []
            output_info = []
            inds_list = []
            inds_orig_list = []
            bands_list = []
            for N_now in unique_N:
                N_now = N_now.item()
                if N_now == 0:  #  or N_now != 1024:
                    continue

                # old_data_1 = self.mgh.data_shaped[0][0][11].copy()
                # checkit = np.where(new_state.supplimental[:]["overall_inds"] == 0)

                # TODO; check the maximum allowable band
                keep = (
                    (band_indices % units == remainder)
                    # & (temp_indices == 0)  #  & (walker_indices == 2)  #  & (band_indices < 50)
                    & (self.band_N_vals[band_indices] == N_now)
                    & (band_indices < len(self.band_edges) - 2)
                    # & (band_indices == 1)
                )  #  & (band_indices == 501) #  # & (N_vals_in <= 256) & (temp_inds == checkit[0].item()) & (walker_inds == checkit[1].item()) #    & (band_indices < 540)  #  &  (temp_inds == 0) & (walker_inds == 0)

                # if self.time > 0:
                #     keep[np.where(keep)[0][1:]] = False
                """if self.time > 0:
                    keep = (
                        (band_indices % units == remainder)
                        & (band_indices == 346) & (walker_indices == 9) & (temp_indices == 0)
                        & (self.band_N_vals[band_indices] == N_now)
                        & (band_indices < len(self.band_edges) - 2)
                    )  #  & (band_indices == 501) #  # & (N_vals_in <= 256) & (temp_inds == checkit[0].item()) & (walker_inds == checkit[1].item()) #    & (band_indices < 540)  #  &  (temp_inds == 0) & (walker_inds == 0)
                    if np.any(keep):
                        breakpoint()"""

                if keep.sum().item() == 0:
                    continue

                total_keep += keep.sum().item()
                # for testing
                # keep[:] = False
                # keep[tmp_keep] = True
                # keep[3000:3020:1] = True

                # permutate for rj just in case
                permute_inds = xp.random.permutation(xp.arange(keep.sum()))
                params_curr = points_curr[keep][permute_inds]
                start_inds_here = start_inds_all[keep][permute_inds]
                inds_here = gb_inds_in[keep][permute_inds]
                factors_here = factors[keep][permute_inds]

                prior_all_curr_here = self.gpu_priors["gb"].logpdf(params_curr, psds=self.mgh.psd_shaped[0][0], walker_inds=group_temp_finder[1][gb_inds][keep][permute_inds])
                if not self.is_rj_prop:
                    assert xp.all(~xp.isinf(prior_all_curr_here))

                temp_inds_here = temp_indices[keep][permute_inds]
                walker_inds_here = walker_indices[keep][permute_inds]
                leaf_inds_here = leaf_indices[keep][permute_inds]
                band_inds_here = band_indices[keep][permute_inds]

                special_band_inds = (temp_inds_here * nwalkers + walker_inds_here) * int(1e6) + band_inds_here

                sort_special = xp.argsort(special_band_inds)

                params_curr = params_curr[sort_special]
                inds_here = inds_here[sort_special]
                factors_here = factors_here[sort_special]
                start_inds_here = start_inds_here[sort_special]
                prior_all_curr_here = prior_all_curr_here[sort_special]
                temp_inds_here = temp_inds_here[sort_special]
                walker_inds_here = walker_inds_here[sort_special]
                leaf_inds_here = leaf_inds_here[sort_special]
                band_inds_here = band_inds_here[sort_special]
                special_band_inds = special_band_inds[sort_special]

                # assert np.all(np.sort(special_band_inds_sorted_here) == special_band_inds_sorted_here)
                
                (
                    uni_special_band_inds_here,
                    uni_index_special_band_inds_here,
                    uni_count_special_band_inds_here,
                ) = self.xp.unique(
                    special_band_inds, return_index=True, return_counts=True
                )
                
                band_start_bin_ind_here = uni_index_special_band_inds_here.astype(
                    np.int32
                )

                band_num_bins_here = uni_count_special_band_inds_here.astype(np.int32)

                band_inds = band_inds_here[uni_index_special_band_inds_here]
                band_temps_inds = temp_inds_here[uni_index_special_band_inds_here]
                band_walkers_inds = walker_inds_here[uni_index_special_band_inds_here]
                
                band_inv_temp_vals_here = band_temps[band_inds, band_temps_inds]

                indiv_info.append((temp_inds_here, walker_inds_here, leaf_inds_here))

                band_bookkeep_info.append(
                    (band_temps_inds, band_walkers_inds, band_inds)
                )

                data_index_here = (((band_inds + 1) % 2) * nwalkers + band_walkers_inds).astype(np.int32)
                noise_index_here = (band_walkers_inds).astype(np.int32)

                # for updates
                update_data_index_here = (((band_inds + 0) % 2) * nwalkers + band_walkers_inds).astype(np.int32)
                
                L_contribution_here = xp.zeros_like(band_inds, dtype=complex)
                p_contribution_here = xp.zeros_like(band_inds, dtype=complex)

                buffer = 5  # bins

                start_band_index = ((self.band_edges[band_inds] / self.df).astype(int) - N_now - 1).astype(np.int32)
                end_band_index = (((self.band_edges[band_inds + 1]) / self.df).astype(int) + N_now + 1).astype(np.int32)
                
                start_band_index[start_band_index < 0] = 0
                end_band_index[end_band_index >= len(self.fd)] = len(self.fd) - 1
                
                band_lengths = end_band_index - start_band_index

                start_interest_band_index = ((self.band_edges[band_inds] / self.df).astype(int)).astype(np.int32)  #  - N_now).astype(np.int32)
                end_interest_band_index = (((self.band_edges[band_inds + 1]) / self.df).astype(int)).astype(np.int32)  #  + N_now).astype(np.int32)
                
                start_interest_band_index[start_interest_band_index < 0] = 0
                end_interest_band_index[end_interest_band_index >= len(self.fd)] = len(self.fd) - 1
                
                band_interest_lengths = end_interest_band_index - start_interest_band_index

                if False:  # self.is_rj_prop:
                    fmin_allow = ((self.band_edges[band_inds] / self.df).astype(int) + 1) * self.df
                    fmax_allow = ((self.band_edges[band_inds + 1] / self.df).astype(int) - 1) * self.df
                    
                else:
                    # proposal limits in a band
                    fmin_allow = ((self.band_edges[band_inds] / self.df).astype(int)- (N_now / 2)) * self.df
                    fmax_allow = ((self.band_edges[band_inds + 1] / self.df).astype(int) + (N_now / 2 )) * self.df

                # if self.is_rj_prop:
                #     breakpoint()
                fmin_allow[fmin_allow < self.band_edges[0]] = self.band_edges[0]
                fmax_allow[fmax_allow > self.band_edges[-1]] = self.band_edges[-1]

                max_data_store_size = band_lengths.max().item()

                num_bands_here = len(band_inds)

                # makes in-model effectively not tempered 
                # if not self.is_rj_prop:
                #    band_inv_temp_vals_here[:] = 1.0

                params_curr_separated = tuple([params_curr[:, i].copy() for i in range(params_curr.shape[1])])
                params_curr_separated_orig = tuple([params_curr[:, i].copy() for i in range(params_curr.shape[1])])
                params_extra_params = (
                    self.waveform_kwargs["T"],
                    self.waveform_kwargs["dt"],
                    N_now,
                    params_curr.shape[0],
                    self.start_freq_ind, 
                    Soms_d_all["sangria"] ** (1/2),
                    Sa_a_all["sangria"] ** (1/2),
                    1e-100, 0.0, 0.0, 0.0, 0.0,  # foreground params for snr -> amp transform
                )

                gb_params_curr_in = self.gb.pyGalacticBinaryParams(
                    *(params_curr_separated + params_curr_separated_orig + params_extra_params)
                )
                
                current_parameter_arrays.append(params_curr_separated)

                data_package = self.gb.pyDataPackage(
                    data[0][0],
                    data[1][0],
                    base_data[0][0],
                    base_data[1][0],
                    psd[0][0],
                    psd[1][0],
                    lisasens[0][0],
                    lisasens[1][0],
                    self.df,
                    self.data_length,
                    self.nwalkers * self.ntemps,
                    self.nwalkers * self.ntemps
                )

                loc_band_index = xp.arange(num_bands_here, dtype=xp.int32)
                band_package = self.gb.pyBandPackage(
                    loc_band_index,
                    data_index_here,
                    noise_index_here,
                    band_start_bin_ind_here,  # uni_index
                    band_num_bins_here,  # uni_count
                    start_band_index,
                    band_lengths,
                    start_interest_band_index,
                    band_interest_lengths,
                    num_bands_here,
                    max_data_store_size,
                    fmin_allow,
                    fmax_allow,
                    update_data_index_here,
                    self.ntemps,
                    band_inds.astype(np.int32),
                    band_walkers_inds.astype(np.int32),
                    band_temps_inds.astype(np.int32),
                    xp.zeros_like(band_temps_inds, dtype=np.int32),  # swaps propsoed
                    xp.zeros_like(band_temps_inds, dtype=np.int32)     # swaps accepted
                )

                accepted_out_here = xp.zeros_like(band_start_bin_ind_here, dtype=xp.int32)

                mcmc_info = self.gb.pyMCMCInfo(
                    L_contribution_here,
                    p_contribution_here,
                    prior_all_curr_here,
                    accepted_out_here,
                    band_inv_temp_vals_here,  # band_inv_temp_vals
                    self.is_rj_prop,
                    self.phase_maximize,
                    self.snr_lim,
                )

                if not self.is_rj_prop:
                    num_proposals_here = self.num_repeat_proposals
                    
                else:
                    num_proposals_here = 1

                num_proposals_per_band = band_num_bins_here * num_proposals_here

                assert start_inds_here.dtype == np.int32

                proposal_info = self.gb.pyStretchProposalPackage(
                    *(self.stretch_friends_args_in + (self.nfriends, len(self.stretch_friends_args_in[0]), num_proposals_here, self.a, ndim, inds_here, factors_here, start_inds_here))
                )

                output_info.append([L_contribution_here, p_contribution_here, accepted_out_here, num_proposals_per_band])

                inputs_now = (
                    data_package,
                    band_package,
                    gb_params_curr_in,
                    mcmc_info,
                    self.gpu_cuda_priors,
                    proposal_info,
                    self.gpu_cuda_wrap,
                    device,
                    do_synchronize,
                )

                # if self.is_rj_prop:
                #     prior_check = xp.zeros(params_curr.shape[0])
                #     self.gb.check_prior_vals(prior_check, self.gpu_cuda_priors, gb_params_curr_in, 100)
                #     prior_check2 = self.gpu_priors["gb"].logpdf(params_curr)
                #     breakpoint()


                N_vals_list.append(N_now)
                inds_list.append(inds_here)
                inds_orig_list.append(inds_here.copy())
                bands_list.append(band_inds_here)
                all_inputs.append(inputs_now)
                st = time.perf_counter()
                # print(params_curr_separated[0].shape[0])
                # if self.is_rj_prop:
                # if self.time > 0:
                #     breakpoint()

                # tmp_check = self.mgh.channel1_data[0][11 * self.data_length + 3911].real + self.mgh.channel1_data[0][29 * self.data_length + 3911].real - self.mgh.channel1_base_data[0][11 * self.data_length + 3911].real
                # print(f"BEFORE {tmp_check}, {self.mgh.channel1_data[0][11 * self.data_length + 3911].real} , {self.mgh.channel1_data[0][29 * self.data_length + 3911].real} , {self.mgh.channel1_base_data[0][11 * self.data_length + 3911].real} ")
                # print(f"BEFORE2 {params_curr_separated[1].min().item()} ")
                self.gb.SharedMemoryMakeNewMove_wrap(*inputs_now)
                self.xp.cuda.runtime.deviceSynchronize()
                # tmp_check = self.mgh.channel1_data[0][11 * self.data_length + 3911].real + self.mgh.channel1_data[0][29 * self.data_length + 3911].real - self.mgh.channel1_base_data[0][11 * self.data_length + 3911].real
                # print(f"After {tmp_check}, {self.mgh.channel1_data[0][11 * self.data_length + 3911].real} , {self.mgh.channel1_data[0][29 * self.data_length + 3911].real} , {self.mgh.channel1_base_data[0][11 * self.data_length + 3911].real} ")
                # print(f"After2 {params_curr_separated[1].min().item()} ")
                
                et = time.perf_counter()
                # print(et - st, N_now)
                # breakpoint()
                
            self.xp.cuda.runtime.deviceSynchronize()
            new_point_info = []
            ll_diff_info = []
            for (
                band_info,
                indiv_info_now,
                N_now,
                outputs,
                current_parameters,
                inds,
                inds_prev,
                bands
            ) in zip(
                band_bookkeep_info, indiv_info, N_vals_list, output_info, current_parameter_arrays, inds_list, inds_orig_list, bands_list
            ):
                ll_contrib_now = outputs[0]
                lp_contrib_now = outputs[1]
                accepted_now = outputs[2]
                num_proposals_now = outputs[3]
            
                per_walker_band_proposals[band_info] += num_proposals_now
                per_walker_band_accepted[band_info] += accepted_now

                # remove accepted
                # print(accepted_now.sum(0) / accepted_now.shape[0])
                temp_tmp, walker_tmp, leaf_tmp = (
                    indiv_info_now[0],
                    indiv_info_now[1],
                    indiv_info_now[2],
                )

                # updates related to newly added sources
                if self.is_rj_prop:
                    gb_inds_orig_check = gb_inds_orig.copy()
                    gb_coords[(temp_tmp[inds], walker_tmp[inds], leaf_tmp[inds])] = xp.asarray(current_parameters).T[inds]

                    gb_inds_orig[(temp_tmp, walker_tmp, leaf_tmp)] = inds
                    new_state.branches_supplimental["gb"].holder["N_vals"][
                        (temp_tmp[inds].get(), walker_tmp[inds].get(), leaf_tmp[inds].get())
                    ] = N_now
                    new_state.branches_supplimental["gb"].holder["band_inds"][
                        (temp_tmp[inds].get(), walker_tmp[inds].get(), leaf_tmp[inds].get())
                    ] = bands[inds].get()

                else:
                    gb_coords[(temp_tmp, walker_tmp, leaf_tmp)] = xp.asarray(current_parameters).T

                    new_band_inds = self.xp.searchsorted(self.band_edges, xp.asarray(current_parameters).T[:, 1] / 1e3, side="right") - 1

                    # if np.any(new_band_inds != bands):
                    #     print(new_band_inds, bands)
                ll_change = self.xp.zeros((ntemps, nwalkers, len(self.band_edges)))
                lp_change = self.xp.zeros((ntemps, nwalkers, len(self.band_edges)))

                self.xp.cuda.runtime.deviceSynchronize()

                ll_change[band_info] = ll_contrib_now

                ll_diff_info.append(ll_change.copy())
                
                ll_adjustment = ll_change.sum(axis=-1)
                log_like_tmp += ll_adjustment

                self.xp.cuda.runtime.deviceSynchronize()

                lp_change[band_info] = lp_contrib_now

                lp_adjustment = lp_change.sum(axis=-1)
                log_prior_tmp += lp_adjustment

            self.xp.cuda.runtime.deviceSynchronize()
            if True:  # self.time > 0:
                ll_after = (
                    self.mgh.get_ll(include_psd_info=True)
                )
                # print(np.abs(new_state.log_like - ll_after).max())
                store_max_diff = np.abs(log_like_tmp[0].get() - ll_after).max()
                # print("CHECKING in:", tmp, store_max_diff)
                if store_max_diff > 3e-4:
                    print("LARGER ERROR:", store_max_diff)
                    self.check_ll_inject(new_state)
                    # self.mgh.get_ll(include_psd_info=True, stop=True)

        new_state.branches["gb"].coords[:] = gb_coords.get()
        if self.is_rj_prop:
            new_state.branches["gb"].inds[:] = gb_inds_orig.get()
        new_state.log_like[:] = log_like_tmp.get()
        new_state.log_prior[:] = log_prior_tmp.get()

        # get updated bands inds ONLY FOR COLD CHAIN and 
        # propogate changes to higher temperatures
        new_freqs = gb_coords[gb_inds_orig, 1]
        new_band_inds = (xp.searchsorted(self.band_edges, new_freqs / 1e3, side="right") - 1)
        new_state.branches["gb"].branch_supplimental.holder["band_inds"][gb_inds_orig.get()] = new_band_inds.get()

        ll_after = (
            self.mgh.get_ll(include_psd_info=True)
        )
        # print(np.abs(new_state.log_like - ll_after).max())
        store_max_diff = np.abs(new_state.log_like[0] - ll_after).max()
        # print("CHECKING 1:", store_max_diff, self.is_rj_prop)
        # if self.time > 0:
        #     breakpoint()
        #     self.check_ll_inject(new_state)

        
        if not self.is_rj_prop:
            # if self.time > 0:
            #     breakpoint()
            # check2 = self.mgh.get_ll()
            old_band_inds_cold_chain = state.branches["gb"].branch_supplimental.holder["band_inds"][0] * state.branches["gb"].inds[0]
            new_band_inds_cold_chain = new_state.branches["gb"].branch_supplimental.holder["band_inds"][0] * state.branches["gb"].inds[0]
            inds_band_change_cold_chain = np.where(new_band_inds_cold_chain != old_band_inds_cold_chain)
            # when adjusting temperatures, be careful here
            if len(inds_band_change_cold_chain[0]) > 0:
                check2 = self.mgh.get_ll(include_psd_info=True)
                # print("SWITCH", len(inds_band_change_cold_chain[0]))
                walker_inds_change_cold_chain = np.tile(inds_band_change_cold_chain[0], (self.ntemps - 1, 1)).flatten()
                old_leaf_inds_change_cold_chain = np.tile(inds_band_change_cold_chain[1], (self.ntemps - 1, 1)).flatten()
                new_temp_inds_change_cold_chain = np.repeat(np.arange(1, self.ntemps), len(inds_band_change_cold_chain[0]))

                special_check = new_temp_inds_change_cold_chain * self.nwalkers + walker_inds_change_cold_chain

                uni_special_check, uni_special_check_count = np.unique(special_check, return_counts=True)

                # get new leaf positions
                temp_leaves = np.ones_like(group_temp_finder[2].reshape(self.ntemps * self.nwalkers, -1).get()[uni_special_check], dtype=int) * (~new_state.branches["gb"].inds.reshape(self.ntemps * self.nwalkers, -1)[uni_special_check])
                temp_leaves_2 = np.cumsum(temp_leaves, axis=-1)
                temp_leaves_2[new_state.branches["gb"].inds.reshape(self.ntemps * self.nwalkers, -1)[uni_special_check]] = -1
                
                leaf_guide_here = np.tile(np.arange(nleaves_max), (len(uni_special_check), 1))
                new_leaf_inds_change_cold_chain = leaf_guide_here[((temp_leaves_2 >= 0) & (temp_leaves_2 <= uni_special_check_count[:, None]))]
                try:
                    assert np.all(~new_state.branches["gb"].inds[new_temp_inds_change_cold_chain, walker_inds_change_cold_chain, new_leaf_inds_change_cold_chain])
                except IndexError:
                    breakpoint()
                    
                new_state.branches["gb"].inds[new_temp_inds_change_cold_chain, walker_inds_change_cold_chain, new_leaf_inds_change_cold_chain] = True
                
                new_state.branches["gb"].coords[new_temp_inds_change_cold_chain, walker_inds_change_cold_chain, new_leaf_inds_change_cold_chain] = new_state.branches["gb"].coords[np.zeros_like(walker_inds_change_cold_chain), walker_inds_change_cold_chain, old_leaf_inds_change_cold_chain]
                new_state.branches["gb"].branch_supplimental.holder["band_inds"][new_temp_inds_change_cold_chain, walker_inds_change_cold_chain, new_leaf_inds_change_cold_chain] = new_state.branches["gb"].branch_supplimental.holder["band_inds"][np.zeros_like(walker_inds_change_cold_chain), walker_inds_change_cold_chain, old_leaf_inds_change_cold_chain]
                new_state.branches["gb"].branch_supplimental.holder["N_vals"][new_temp_inds_change_cold_chain, walker_inds_change_cold_chain, new_leaf_inds_change_cold_chain] = new_state.branches["gb"].branch_supplimental.holder["N_vals"][np.zeros_like(walker_inds_change_cold_chain), walker_inds_change_cold_chain, old_leaf_inds_change_cold_chain]
            
                # adjust data
                adjust_binaries = xp.asarray(new_state.branches["gb"].coords[0, inds_band_change_cold_chain[0], inds_band_change_cold_chain[1]])
                adjust_binaries_in = self.parameter_transforms.both_transforms(
                    adjust_binaries, xp=xp
                )
                adjust_walker_inds = inds_band_change_cold_chain[0]
                adjust_band_new = new_state.branches["gb"].branch_supplimental.holder["band_inds"][0, inds_band_change_cold_chain[0], inds_band_change_cold_chain[1]]
                adjust_band_old = state.branches["gb"].branch_supplimental.holder["band_inds"][0, inds_band_change_cold_chain[0], inds_band_change_cold_chain[1]]
                # N_vals_in = new_state.branches["gb"].branch_supplimental.holder["N_vals"][0, inds_band_change_cold_chain[0], inds_band_change_cold_chain[1]]
                
                adjust_binaries_in_in = xp.concatenate([
                    adjust_binaries_in,
                    adjust_binaries_in
                ], axis=0)

                data_index = xp.concatenate([
                    xp.asarray(((adjust_band_old + 0) % 2) * nwalkers + adjust_walker_inds),
                    xp.asarray(((adjust_band_new + 0) % 2) * nwalkers + adjust_walker_inds)
                ]).astype(xp.int32)

                N_vals_in_in = xp.concatenate([
                    xp.asarray(self.band_N_vals[adjust_band_old]),
                    xp.asarray(self.band_N_vals[adjust_band_new])
                ])

                factors = xp.concatenate([
                    +xp.ones_like(adjust_band_old, dtype=xp.float64),  # remove
                    -xp.ones_like(adjust_band_old, dtype=xp.float64)  # add
                ])

                self.gb.generate_global_template(
                    adjust_binaries_in_in,
                    data_index,
                    self.mgh.data_list,
                    N=N_vals_in_in,
                    factors=factors,
                    data_length=self.mgh.data_length,
                    data_splits=self.mgh.gpu_splits,
                    **waveform_kwargs_now
                )
                check3 = self.mgh.get_ll(include_psd_info=True)

                # print(check3 - check2)

                if np.any(
                    self.band_N_vals[adjust_band_old] !=
                    self.band_N_vals[adjust_band_new]
                ):
                    walkers_focus = np.unique(inds_band_change_cold_chain[0][
                        self.band_N_vals.get()[adjust_band_old] !=
                        self.band_N_vals.get()[adjust_band_new]
                    ])
                    # print(f"specific change in N across boundary: {walkers_focus}")
                    new_state.log_like[0, walkers_focus] = self.mgh.get_ll(include_psd_info=True)[walkers_focus]

        # self.check_ll_inject(new_state)
        # breakpoint()
        ll_after = (
            self.mgh.get_ll(include_psd_info=True)
        )
        # print(np.abs(new_state.log_like - ll_after).max())
        store_max_diff = np.abs(new_state.log_like[0] - ll_after).max()
        # print("CHECKING 2:", store_max_diff, self.is_rj_prop)

        self.mempool.free_all_blocks()
        # get accepted fraction
        if not self.is_rj_prop:
            accepted_check_tmp = np.zeros_like(
                state.branches_inds["gb"], dtype=bool
            )
            accepted_check_tmp[state.branches_inds["gb"]] = np.all(
                np.abs(
                    new_state.branches_coords["gb"][
                        state.branches_inds["gb"]
                    ]
                    - state.branches_coords["gb"][state.branches_inds["gb"]]
                )
                > 0.0,
                axis=-1,
            )
            proposed = gb_inds.get()
            accepted_check = accepted_check_tmp.sum(
                axis=(1, 2)
            ) / proposed.sum(axis=(1, 2))
        else:
            accepted_check_tmp = (
                new_state.branches_inds["gb"] == (~state.branches_inds["gb"])
            )

            proposed = gb_inds.get()
            accepted_check = accepted_check_tmp.sum(axis=(1, 2)) / proposed.sum(axis=(1, 2))
            
        # manually tell temperatures how real overall acceptance fraction is
        number_of_walkers_for_accepted = np.floor(nwalkers * accepted_check).astype(int)

        accepted_inds = np.tile(np.arange(nwalkers), (ntemps, 1))

        accepted = np.zeros((ntemps, nwalkers), dtype=bool)
        accepted[accepted_inds < number_of_walkers_for_accepted[:, None]] = True

        tmp1 = np.all(
            np.abs(
                new_state.branches_coords["gb"]
                - state.branches_coords["gb"]
            )
            > 0.0,
            axis=-1,
        ).sum(axis=(2,))
        tmp2 = new_state.branches_inds["gb"].sum(axis=(2,))

        # add to move-specific accepted information
        self.accepted += tmp1
        if isinstance(self.num_proposals, int):
            self.num_proposals = tmp2
        else:
            self.num_proposals += tmp2

        new_inds = xp.asarray(new_state.branches_inds["gb"])
            
        # in-model inds will not change
        tmp_freqs_find_bands = xp.asarray(new_state.branches_coords["gb"][:, :, :, 1])

        # calculate current band counts
        band_here = (xp.searchsorted(self.band_edges, tmp_freqs_find_bands.flatten() / 1e3, side="right") - 1).reshape(tmp_freqs_find_bands.shape)

        # get binaries per band
        special_band_here_num_per_band = ((group_temp_finder[0] * nwalkers + group_temp_finder[1]) * int(1e6) + band_here)[new_inds]
        unique_special_band_here_num_per_band, unique_special_band_here_num_per_band_count = xp.unique(special_band_here_num_per_band, return_counts=True)
        temp_walker_index_num_per_band = (unique_special_band_here_num_per_band / 1e6).astype(int)
        temp_index_num_per_band = (temp_walker_index_num_per_band / nwalkers).astype(int)
        walker_index_num_per_band = temp_walker_index_num_per_band - temp_index_num_per_band * nwalkers
        band_index_num_per_band = unique_special_band_here_num_per_band - temp_walker_index_num_per_band * int(1e6)

        per_walker_band_counts = xp.zeros((ntemps, nwalkers, self.num_bands), dtype=int)
        per_walker_band_counts[temp_index_num_per_band, walker_index_num_per_band, band_index_num_per_band] = unique_special_band_here_num_per_band_count
        
        # TEMPERING
        self.temperature_control.swaps_accepted = np.zeros(ntemps - 1)
        self.temperature_control.swaps_proposed = np.zeros(ntemps - 1)

        band_swaps_accepted = np.zeros((len(self.band_edges) - 1, self.ntemps - 1), dtype=int)
        band_swaps_proposed = np.zeros((len(self.band_edges) - 1, self.ntemps - 1), dtype=int)
        current_band_counts = np.zeros((len(self.band_edges) - 1, self.ntemps), dtype=int)
        
        # if self.is_rj_prop:
        #     print("1st count check:", new_state.branches["gb"].inds.sum(axis=-1).mean(axis=-1), "\nll:", new_state.log_like[0] - orig_store, new_state.log_like[0])
        
        # if self.time > 0:
        #     self.check_ll_inject(new_state)
        if (
            self.temperature_control is not None
            and self.time % 1 == 0
            and self.ntemps > 1
            and self.is_rj_prop
            # and False
        ):

            new_coords_after_tempering = xp.asarray(np.zeros_like(new_state.branches["gb"].coords))
            new_inds_after_tempering = xp.asarray(np.zeros_like(new_state.branches["gb"].inds))

            # TODO: check if N changes / need to deal with that
            betas = self.temperature_control.betas

            # cannot find them yourself because of higher temps moving across band edge / need supplimental band inds
            band_inds_temp = new_state.branches["gb"].branch_supplimental.holder["band_inds"][new_state.branches["gb"].inds]
            temp_inds_temp = group_temp_finder[0].get()[new_state.branches["gb"].inds]
            walker_inds_temp = group_temp_finder[1].get()[new_state.branches["gb"].inds]

            bands_guide = np.tile(np.arange(self.num_bands), (self.ntemps, self.nwalkers, 1)).transpose(2, 1, 0).flatten()
            temps_guide = np.repeat(np.arange(self.ntemps)[:, None], self.nwalkers * self.num_bands).reshape(self.ntemps, self.nwalkers, self.num_bands).transpose(2, 1, 0).flatten()
            walkers_guide = np.repeat(np.arange(self.nwalkers)[:, None], self.ntemps * self.num_bands).reshape(self.nwalkers, self.ntemps, self.num_bands).transpose(2, 0, 1).flatten()
            
            walkers_permuted = np.asarray([np.random.permutation(np.arange(self.nwalkers)) for _ in range(self.ntemps * self.num_bands)]).reshape(self.num_bands, self.ntemps, self.nwalkers).transpose(0, 2, 1).flatten()

            # special_inds_guide = bands_guide * int(1e6) + walkers_permuted * int(1e3) + temps_guide

            coords_in = new_state.branches["gb"].coords[new_state.branches["gb"].inds]
            
            # N_vals_in = new_state.branches["gb"].branch_supplimental.holder["N_vals"][new_state.branches["gb"].inds]
            unique_N = np.unique(self.band_N_vals).get()
            # remove 0
            unique_N = unique_N[unique_N != 0]

            bands_in = band_inds_temp
            temps_in = temp_inds_temp
            walkers_in = walker_inds_temp

            main_gpu = self.xp.cuda.runtime.getDevice()

            walkers_info = walkers_permuted if self.temperature_control.permute else walkers_guide
            units = 2
            start_unit = np.random.randint(0, units)
            for unit in range(units):
                current_band_remainder = (start_unit + unit) % units
                for N_now in unique_N:
                    keep = (bands_in % units == current_band_remainder) & (self.band_N_vals[xp.asarray(bands_in)].get() == N_now)
                    keep_base = (bands_guide % units == current_band_remainder) & (self.band_N_vals[xp.asarray(bands_guide)].get() == N_now)
                    bands_in_tmp = bands_in[keep]
                    temps_in_tmp = temps_in[keep]
                    walkers_in_tmp = walkers_in[keep]
                    coords_in_tmp = coords_in[keep]

                    # adjust for N_val
                    special_inds_bins = bands_in_tmp * int(1e8) + walkers_in_tmp * int(1e4) + temps_in_tmp

                    bands_guide_keep = bands_guide[keep_base]
                    temps_guide_keep = temps_guide[keep_base]
                    walkers_info_keep = walkers_info[keep_base]

                    special_inds_guide_keep = bands_guide_keep * int(1e8) + walkers_info_keep * int(1e4) + temps_guide_keep
                    sort_tmp = np.argsort(special_inds_guide_keep)
                    sorted_special_inds_guide_keep = special_inds_guide_keep[sort_tmp]
                    sorting_info_need_to_revert = np.searchsorted(sorted_special_inds_guide_keep, special_inds_bins, side="left")
                    
                    sorting_info = sort_tmp[sorting_info_need_to_revert]

                    assert np.all(special_inds_guide_keep[sorting_info] == special_inds_bins)

                    sort_bins = np.argsort(sorting_info)

                    # p1 = self.mgh.psd_shaped[0][0][8, 19748:20132]
                    # p2 = self.mgh.psd_shaped[1][0][8, 19748:20132]
                    # c2 =  self.mgh.data_shaped[1][0][26, 19748:20132] + self.mgh.data_shaped[1][0][8, 19748:20132]- self.mgh.channel2_base_data[0][8 * self.data_length + 19748:8 * self.data_length + 20132]
                    # c1 =  self.mgh.data_shaped[0][0][26, 19748:20132] + self.mgh.data_shaped[0][0][8, 19748:20132]- self.mgh.channel1_base_data[0][8 * self.data_length + 19748:8 * self.data_length + 20132]
                    # args_sorting = np.arange(len(sorting_info))  # np.argsort(sorting_info)
                    bands_in_tmp = bands_in_tmp[sort_bins]
                    temps_in_tmp = temps_in_tmp[sort_bins]
                    walkers_in_tmp = walkers_in_tmp[sort_bins]
                    coords_in_tmp = coords_in_tmp[sort_bins]
                    sorting_info = sorting_info[sort_bins]

                    uni_sorted, uni_index, uni_inverse, uni_counts = xp.unique(xp.asarray(sorting_info), return_index=True, return_counts=True, return_inverse=True)
                    
                    band_info_start_index_bin = xp.zeros_like(bands_guide_keep, dtype=np.int32)
                    band_info_num_bin = xp.zeros_like(bands_guide_keep, dtype=np.int32)

                    band_info_start_index_bin[uni_sorted.astype(np.int32)] = uni_index.astype(np.int32)
                    band_info_num_bin[uni_sorted.astype(np.int32)] = uni_counts.astype(np.int32)

                    params_curr_separated = tuple([xp.asarray(coords_in_tmp[:, i].copy()) for i in range(coords_in_tmp.shape[1])])
                    params_curr_separated_orig = tuple([xp.asarray(coords_in_tmp[:, i].copy()) for i in range(coords_in_tmp.shape[1])])
                    params_extra_params = (
                        self.waveform_kwargs["T"],
                        self.waveform_kwargs["dt"],
                        N_now,
                        coords_in.shape[0],
                        self.start_freq_ind, 
                        Soms_d_all["sangria"] ** (1/2),
                        Sa_a_all["sangria"] ** (1/2),
                        1e-100, 0.0, 0.0, 0.0, 0.0,  # foreground params for snr -> amp transform
                    )

                    gb_params_curr_in = self.gb.pyGalacticBinaryParams(
                        *(params_curr_separated + params_curr_separated_orig + params_extra_params)
                    )

                    # current_parameter_arrays.append(params_curr_separated)

                    data_package = self.gb.pyDataPackage(
                        data[0][0],
                        data[1][0],
                        base_data[0][0],
                        base_data[1][0],
                        psd[0][0],
                        psd[1][0],
                        lisasens[0][0],
                        lisasens[1][0],
                        self.df,
                        self.data_length,
                        self.nwalkers * self.ntemps,
                        self.nwalkers * self.ntemps
                    )

                    data_index_here = xp.asarray((((bands_guide_keep + 1) % 2) * self.nwalkers + walkers_info_keep)).astype(np.int32)
                    noise_index_here = xp.asarray(walkers_info_keep.copy()).astype(np.int32)
                    update_data_index_here = xp.asarray((((bands_guide_keep + 0) % 2) * self.nwalkers + walkers_info_keep)).astype(np.int32)
                    
                    start_band_index = ((self.band_edges[bands_guide_keep] / self.df).astype(int) - N_now - 1).astype(np.int32)
                    end_band_index = (((self.band_edges[bands_guide_keep + 1]) / self.df).astype(int) + N_now + 1).astype(np.int32)
                    
                    start_band_index[start_band_index < 0] = 0
                    end_band_index[end_band_index >= len(self.fd)] = len(self.fd) - 1
                    
                    band_lengths = end_band_index - start_band_index

                    start_interest_band_index = ((self.band_edges[bands_guide_keep] / self.df).astype(int)).astype(np.int32)  #  - N_now).astype(np.int32)
                    end_interest_band_index = (((self.band_edges[bands_guide_keep + 1]) / self.df).astype(int)).astype(np.int32)  #  + N_now).astype(np.int32)
                    
                    start_interest_band_index[start_interest_band_index < 0] = 0
                    end_interest_band_index[end_interest_band_index >= len(self.fd)] = len(self.fd) - 1
                    
                    band_interest_lengths = end_interest_band_index - start_interest_band_index


                    # proposal limits in a band
                    fmin_allow = xp.asarray(((self.band_edges[bands_guide_keep] / self.df).astype(int) - (N_now / 2))) * self.df
                    fmax_allow = xp.asarray(((self.band_edges[bands_guide_keep + 1] / self.df).astype(int) + (N_now / 2))) * self.df

                    fmin_allow[fmin_allow < self.band_edges[0]] = self.band_edges[0]
                    fmax_allow[fmax_allow > self.band_edges[-1]] = self.band_edges[-1]

                    max_data_store_size = band_lengths.max().item()

                    num_bands_here = len(band_info_start_index_bin)

                    num_swap_setups = np.unique(bands_guide_keep).shape[0] * self.nwalkers

                    swaps_proposed_here = xp.zeros(num_swap_setups * (self.ntemps - 1), dtype=np.int32)
                    swaps_accepted_here = xp.zeros(num_swap_setups * (self.ntemps - 1), dtype=np.int32)
                    
                    bands_guide_keep = xp.asarray(bands_guide_keep).astype(np.int32)
                    walkers_info_keep = xp.asarray(walkers_info_keep).astype(np.int32)
                    temps_guide_keep = xp.asarray(temps_guide_keep).astype(np.int32)
                    loc_band_index = xp.arange(num_bands_here, dtype=xp.int32)
                    
                    before_loc_band_index = loc_band_index.copy()
                    before_band_info_start_index_bin = band_info_start_index_bin.copy()
                    before_band_info_num_bin = band_info_num_bin.copy()

                    band_package = self.gb.pyBandPackage(
                        loc_band_index,
                        data_index_here,
                        noise_index_here,
                        band_info_start_index_bin,  # uni_index
                        band_info_num_bin,  # uni_count
                        start_band_index,
                        band_lengths,
                        start_interest_band_index,
                        band_interest_lengths,
                        num_bands_here,
                        max_data_store_size,
                        fmin_allow,
                        fmax_allow,
                        update_data_index_here,
                        self.ntemps,
                        bands_guide_keep, 
                        walkers_info_keep,
                        temps_guide_keep,
                        swaps_proposed_here,
                        swaps_accepted_here
                    )

                    band_inv_temp_vals_here = band_temps[bands_guide_keep, temps_guide_keep]

                    accepted_out_here = xp.zeros_like(bands_guide_keep, dtype=xp.int32)
                    L_contribution_here = xp.zeros_like(bands_guide_keep, dtype=complex)
                    p_contribution_here = xp.zeros_like(bands_guide_keep, dtype=complex)
                    prior_all_curr_here = xp.zeros_like(bands_guide_keep, dtype=np.float64)
                    
                    mcmc_info = self.gb.pyMCMCInfo(
                        L_contribution_here,
                        p_contribution_here,
                        prior_all_curr_here,
                        accepted_out_here,
                        band_inv_temp_vals_here,  # band_inv_temp_vals
                        self.is_rj_prop,
                        False,  # phased maximize
                        self.snr_lim,
                    )

                    inds_here = xp.ones_like(temps_in_tmp, dtype=bool)
                    factors_here = xp.zeros_like(temps_in_tmp, dtype=float)
                    start_inds_here = xp.zeros_like(temps_in_tmp, dtype=np.int32)
                    
                    num_proposals_here = 1
                    
                    assert start_inds_here.dtype == np.int32
                    proposal_info = self.gb.pyStretchProposalPackage(
                        *(self.stretch_friends_args_in + (self.nfriends, len(self.stretch_friends_args_in[0]), num_proposals_here, self.a, ndim, inds_here, factors_here, start_inds_here))
                    )

                    inputs_now = (
                        data_package,
                        band_package,
                        gb_params_curr_in,
                        mcmc_info,
                        self.gpu_cuda_priors,
                        proposal_info,
                        self.gpu_cuda_wrap,
                        num_swap_setups,
                        device,
                        do_synchronize,
                        -1, 
                        100000
                    )

                    self.gb.SharedMemoryMakeTemperingMove_wrap(*inputs_now)

                    self.xp.cuda.runtime.deviceSynchronize()

                    walkers_info_keep_per_bin = xp.repeat(walkers_info_keep, list(band_info_num_bin.get()))
                    temps_guide_keep_per_bin = xp.repeat(temps_guide_keep, list(band_info_num_bin.get()))
                    start_bin_ind_mapping = xp.repeat(band_info_start_index_bin, list(band_info_num_bin.get()))
                    uni_after, uni_after_start_index, uni_after_inverse = xp.unique(start_bin_ind_mapping, return_index=True, return_inverse=True)
                    ind_map = xp.arange(len(start_bin_ind_mapping)) - xp.arange(len(start_bin_ind_mapping))[uni_after_start_index][uni_after_inverse] + start_bin_ind_mapping
                    
                    # out = []
                    # for i in range(len(band_info_num_bin_has)):
                    #     num_bins = band_info_num_bin_has[i].item()
                    #     start_index = band_info_start_index_bin_has[i].item()
                    #     for j in range(num_bins):
                    #         out.append(start_index + j)

                    # breakpoint()
                    new_coords_mapped = coords_in_tmp[ind_map.get()]

                    special_after_tempering = temps_guide_keep_per_bin * self.nwalkers + walkers_info_keep_per_bin
                    
                    sorted_after = xp.argsort(special_after_tempering)
                    special_after_tempering = special_after_tempering[sorted_after]
                    walkers_info_keep_per_bin = walkers_info_keep_per_bin[sorted_after]
                    temps_guide_keep_per_bin = temps_guide_keep_per_bin[sorted_after]
                    new_coords_mapped = new_coords_mapped[sorted_after.get()]

                    uni_after, uni_index_after, uni_inverse_after = xp.unique(special_after_tempering, return_index=True, return_inverse=True)

                    relative_leaf_info_after_tempering = xp.arange(len(special_after_tempering)) - uni_index_after[uni_inverse_after]

                    leaf_start_per_walker = new_inds_after_tempering.argmin(axis=-1)[temps_guide_keep_per_bin[uni_index_after], walkers_info_keep_per_bin[uni_index_after]]
                    
                    absolute_leaf_info_after_tempering = leaf_start_per_walker[uni_inverse_after] + relative_leaf_info_after_tempering

                    assert (np.all(~new_inds_after_tempering[temps_guide_keep_per_bin, walkers_info_keep_per_bin, absolute_leaf_info_after_tempering]))
                    new_coords_after_tempering[temps_guide_keep_per_bin, walkers_info_keep_per_bin, absolute_leaf_info_after_tempering] = new_coords_mapped
                    new_inds_after_tempering[temps_guide_keep_per_bin, walkers_info_keep_per_bin, absolute_leaf_info_after_tempering] = True

                    bands_unique_here = np.unique(bands_guide_keep).get()
                    band_swaps_accepted[bands_unique_here] += swaps_accepted_here.reshape(bands_unique_here.shape[0], self.nwalkers, self.ntemps - 1).sum(axis=1).get()
                    band_swaps_proposed[bands_unique_here] += swaps_proposed_here.reshape(bands_unique_here.shape[0], self.nwalkers, self.ntemps - 1).sum(axis=1).get()

                    ll_change = self.xp.zeros((ntemps, nwalkers, len(self.band_edges)))

                    self.xp.cuda.runtime.deviceSynchronize()

                    ll_change[(temps_guide_keep,walkers_info_keep,bands_guide_keep)] = L_contribution_here
                    
                    ll_adjustment = ll_change.sum(axis=-1)
                    log_like_tmp += ll_adjustment
                    
            new_state.branches["gb"].coords[:] = new_coords_after_tempering.get()
            new_state.branches["gb"].inds[:] = new_inds_after_tempering.get()
            new_state.log_like[:] = log_like_tmp.get()

            new_freqs = new_coords_after_tempering[new_inds_after_tempering, 1]
            new_band_inds = (xp.searchsorted(self.band_edges, new_freqs / 1e3, side="right") - 1)
            new_state.branches["gb"].branch_supplimental.holder["band_inds"][new_inds_after_tempering.get()] = new_band_inds.get()

            # breakpoint()
            # self.check_ll_inject(new_state)
            # breakpoint()
            
            # adjust priors accordingly
            log_prior_new_per_bin = xp.zeros_like(
                new_state.branches_inds["gb"], dtype=xp.float64
            )
            # self.gpu_priors
            log_prior_new_per_bin[
                new_state.branches_inds["gb"]
            ] = self.gpu_priors["gb"].logpdf(
                xp.asarray(
                    new_state.branches_coords["gb"][
                        new_state.branches_inds["gb"]
                    ]
                ),
                psds=self.mgh.psd_shaped[0][0],
                walker_inds=group_temp_finder[1].get()[new_state.branches_inds["gb"]]
            )

            new_state.log_prior = log_prior_new_per_bin.sum(axis=-1).get()
            
            ratios = (band_swaps_accepted / band_swaps_proposed).T #  self.swaps_accepted / self.swaps_proposed
            ratios[np.isnan(ratios)] = 0.0

            # only change those with a binary in them

            # adapt if desired
            if self.time > 50:
                betas0 = band_temps.copy().T.get()
                betas1 = betas0.copy()

                # Modulate temperature adjustments with a hyperbolic decay.
                decay = self.temperature_control.adaptation_lag / (self.time + self.temperature_control.adaptation_lag)
                kappa = decay / self.temperature_control.adaptation_time

                # Construct temperature adjustments.
                dSs = kappa * (ratios[:-1] - ratios[1:])

                # Compute new ladder (hottest and coldest chains don't move).
                deltaTs = np.diff(1 / betas1[:-1], axis=0)

                deltaTs *= np.exp(dSs)
                betas1[1:-1] = 1 / (np.cumsum(deltaTs, axis=0) + 1 / betas1[0])

                # Don't mutate the ladder here; let the client code do that.
                dbetas = betas1 - betas0

                band_temps += self.xp.asarray(dbetas.T)

            # band_temps[:] = band_temps[553, :][None, :]

            # only increase time if it is adaptive.
            new_state.betas = self.temperature_control.betas.copy()
            
            self.mempool.free_all_blocks()
            # print(
            #     self.is_rj_prop,
            #     band_swaps_accepted[350] / band_swaps_proposed[350],
            #     band_swaps_accepted[450] / band_swaps_proposed[450],
            #     band_swaps_accepted[501] / band_swaps_proposed[501]
            # )

        self.mempool.free_all_blocks()
        if self.time % 1 == 0:
            ll_after = (
                self.mgh.get_ll(include_psd_info=True)
            )
            # print(np.abs(new_state.log_like - ll_after).max())
            store_max_diff = np.abs(new_state.log_like[0] - ll_after).max()
            # print("CHECKING:", store_max_diff, self.is_rj_prop)
            if store_max_diff > 1e-5:
                ll_after = (
                    self.mgh.get_ll(include_psd_info=True, stop=True)
                )

                if store_max_diff > 1.0:
                    breakpoint()

                # reset data and fix likelihood
                new_state.log_like[0] = self.check_ll_inject(new_state)
            

        self.time += 1
        # self.xp.cuda.runtime.deviceSynchronize()

        new_state.update_band_information(
            band_temps.get(), per_walker_band_proposals.sum(axis=1).get().T, per_walker_band_accepted.sum(axis=1).get().T, band_swaps_proposed, band_swaps_accepted,
            per_walker_band_counts.get(), self.is_rj_prop
        )
        # TODO: check rj numbers

        # new_state.log_like[:] = self.check_ll_inject(new_state)

        self.mempool.free_all_blocks()

        if self.is_rj_prop:
            pass  # print(self.name, "2nd count check:", new_state.branches["gb"].inds.sum(axis=-1).mean(axis=-1), "\nll:", new_state.log_like[0] - orig_store, new_state.log_like[0])
        return new_state, accepted

    def check_ll_inject(self, new_state):

        check_ll = self.mgh.get_ll(include_psd_info=True).copy()

        nleaves_max = new_state.branches["gb"].shape[-2]
        for i in range(2):
            self.mgh.channel1_data[0][self.nwalkers * self.data_length * i: self.nwalkers * self.data_length * (i + 1)] = self.mgh.channel1_base_data[0][:]
            self.mgh.channel2_data[0][self.nwalkers * self.data_length * i: self.nwalkers * self.data_length * (i + 1)] = self.mgh.channel2_base_data[0][:]
        
        coords_out_gb = new_state.branches["gb"].coords[0, new_state.branches["gb"].inds[0]]
        coords_in_in = self.parameter_transforms.both_transforms(coords_out_gb)

        band_inds = np.searchsorted(self.band_edges.get(), coords_in_in[:, 1], side="right") - 1
        assert np.all(band_inds == new_state.branches["gb"].branch_supplimental.holder["band_inds"][0, new_state.branches["gb"].inds[0]])

        walker_vals = np.tile(np.arange(self.nwalkers), (nleaves_max, 1)).transpose((1, 0))[new_state.branches["gb"].inds[0]]

        data_index_1 = ((band_inds % 2) + 0) * self.nwalkers + walker_vals

        data_index = xp.asarray(data_index_1).astype(xp.int32)

        # goes in as -h
        factors = -xp.ones_like(data_index, dtype=xp.float64)

        waveform_kwargs_tmp = self.waveform_kwargs.copy()

        N_vals = self.band_N_vals[band_inds]
        self.gb.generate_global_template(
            coords_in_in,
            data_index,
            self.mgh.data_list,
            batch_size=1000,
            data_length=self.data_length,
            factors=factors,
            N=N_vals,
            data_splits=self.mgh.gpu_splits,
            **waveform_kwargs_tmp,
        )

        check_ll_new = self.mgh.get_ll(include_psd_info=True)
        check_ll_diff1 = check_ll_new - check_ll
        # print(check_ll_diff1)

        # breakpoint()
        return check_ll_new

        # breakpoint()

        # # print(self.accepted / self.num_proposals)

        # # MULTIPLE TRY after RJ

        # if False: # self.is_rj_prop:
            
        #     inds_added = np.where(new_state.branches["gb"].inds.astype(int) - state.branches["gb"].inds.astype(int) == +1)

        #     new_coords = xp.asarray(new_state.branches["gb"].coords[inds_added])
        #     temp_inds_add = xp.repeat(xp.arange(ntemps)[:, None], nwalkers * nleaves_max, axis=-1).reshape(ntemps, nwalkers, nleaves_max)[inds_added]
        #     walker_inds_add = xp.repeat(xp.arange(nwalkers)[:, None], ntemps * nleaves_max, axis=-1).reshape(nwalkers, ntemps, nleaves_max).transpose(1, 0, 2)[inds_added]
        #     leaf_inds_add = xp.repeat(xp.arange(nleaves_max)[:, None], ntemps * nwalkers, axis=-1).reshape(nleaves_max, ntemps, nwalkers).transpose(1, 2, 0)[inds_added]
        #     band_inds_add = xp.searchsorted(self.band_edges, new_coords[:, 1] / 1e3, side="right") - 1
        #     N_vals_add = xp.asarray(new_state.branches["gb"].branch_supplimental.holder["N_vals"][inds_added])
            
        #     #### RIGHT NOW WE ARE EXPERIMENTING WITH NO OR SMALL CHANGE IN FREQUENCY
        #     # because if it is added in RJ, it has locked on in frequency in some form
            
        #     # randomize order
        #     inds_random = xp.random.permutation(xp.arange(len(new_coords)))
        #     new_coords = new_coords[inds_random]
        #     temp_inds_add = temp_inds_add[inds_random]
        #     walker_inds_add = walker_inds_add[inds_random]
        #     leaf_inds_add = leaf_inds_add[inds_random]
        #     band_inds_add = band_inds_add[inds_random]
        #     N_vals_add = N_vals_add[inds_random]

        #     special_band_map = leaf_inds_add * int(1e12) + walker_inds_add * int(1e6) + band_inds_add
        #     inds_sorted_special = xp.argsort(special_band_map)
        #     special_band_map = special_band_map[inds_sorted_special]
        #     new_coords = new_coords[inds_sorted_special]
        #     temp_inds_add = temp_inds_add[inds_sorted_special]
        #     walker_inds_add = walker_inds_add[inds_sorted_special]
        #     leaf_inds_add = leaf_inds_add[inds_sorted_special]
        #     band_inds_add = band_inds_add[inds_sorted_special]
        #     N_vals_add = N_vals_add[inds_sorted_special]

        #     unique_special_bands, unique_special_bands_index, unique_special_bands_inverse = xp.unique(special_band_map, return_index=True, return_inverse=True)
        #     group_index = xp.arange(len(special_band_map)) - xp.arange(len(special_band_map))[unique_special_bands_index][unique_special_bands_inverse]
            
        #     band_splits = 3
        #     num_try = 1000
        #     for group in range(group_index.max().item()):
        #         for split_i in range(band_splits):
        #             group_split = (group_index == group) & (band_inds_add % band_splits == split_i)
                    
        #             new_coords_group_split = new_coords[group_split]
        #             temp_inds_add_group_split = temp_inds_add[group_split]
        #             walker_inds_add_group_split = walker_inds_add[group_split]
        #             leaf_inds_add_group_split = leaf_inds_add[group_split]
        #             band_inds_add_group_split = band_inds_add[group_split]
        #             N_vals_add_group_split = N_vals_add[group_split]

        #             coords_remove = xp.repeat(new_coords_group_split, num_try, axis=0)
        #             coords_add = coords_remove.copy()

        #             inds_params = xp.array([0, 2, 3, 4, 5, 6, 7])
        #             coords_add[:, inds_params] = self.gpu_priors["gb"].rvs(size=coords_remove.shape[0])[:, inds_params]
                    
        #             log_proposal_pdf = self.gpu_priors["gb"].logpdf(coords_add)
        #             # remove logpdf from fchange for now
        #             log_proposal_pdf -= self.gpu_priors["gb"].priors_in[1].logpdf(coords_add[:, 1])

        #             priors_remove = self.gpu_priors["gb"].logpdf(coords_remove)
        #             priors_add = self.gpu_priors["gb"].logpdf(coords_add)

        #             if xp.any(coords_add[:, 1] != coords_remove[:, 1]):
        #                 raise NotImplementedError("Assumes frequencies are the same.")
        #                 independent = False
        #             else:
        #                 independent = True

        #             coords_remove_in = self.parameter_transforms.both_transforms(coords_remove, xp=xp)
        #             coords_add_in = self.parameter_transforms.both_transforms(coords_add, xp=xp)
                    
        #             waveform_kwargs_tmp = self.waveform_kwargs.copy()
        #             waveform_kwargs_tmp.pop("N")

        #             data_index_in = xp.repeat(temp_inds_add_group_split * nwalkers + walker_inds_add_group_split, num_try).astype(xp.int32)
        #             noise_index_in = data_index_in.copy()
        #             N_vals_add_group_split_in = self.xp.repeat(N_vals_add_group_split, num_try)

        #             ll_diff = self.xp.asarray(self.gb.swap_likelihood_difference(coords_remove_in, coords_add_in, self.mgh.data_list, self.mgh.psd_list, data_index=data_index_in, noise_index=noise_index_in, N=N_vals_add_group_split_in, data_length=self.data_length, data_splits=self.mgh.gpu_splits, **waveform_kwargs_tmp))
                    
        #             ll_diff[self.xp.isnan(ll_diff)] = -1e300

        #             band_inv_temps = self.xp.repeat(band_temps[(band_inds_add_group_split, temp_inds_add_group_split)], num_try)
                    
        #             logP = band_inv_temps * ll_diff + priors_add

        #             from eryn.moves.multipletry import logsumexp, get_mt_computations

        #             band_inv_temps = band_inv_temps.reshape(-1, num_try)
        #             ll_diff = ll_diff.reshape(-1, num_try)
        #             logP = logP.reshape(-1, num_try)
        #             priors_add = priors_add.reshape(-1, num_try)
        #             log_proposal_pdf = log_proposal_pdf.reshape(-1, num_try)
        #             coords_add = coords_add.reshape(-1, num_try, ndim)
                    
        #             log_importance_weights, log_sum_weights, inds_group_split = get_mt_computations(logP, log_proposal_pdf, symmetric=False, xp=self.xp)
                    
        #             inds_tuple = (self.xp.arange(len(inds_group_split)), inds_group_split)

        #             ll_diff_out = ll_diff[inds_tuple]
        #             logP_out = logP[inds_tuple]
        #             priors_add_out = priors_add[inds_tuple]
        #             coords_add_out = coords_add[inds_tuple]
        #             log_proposal_pdf_out = log_proposal_pdf[inds_tuple]

        #             if not independent:
        #                 raise NotImplementedError
        #             else:
        #                 aux_coords_add = coords_add.copy()
        #                 aux_ll_diff = ll_diff.copy()
        #                 aux_priors_add = priors_add.copy()
        #                 aux_log_proposal_pdf = log_proposal_pdf.copy()

        #                 aux_coords_add[:, 0] = new_coords_group_split
        #                 aux_ll_diff[:, 0] = 0.0  # diff is zero because the points are already in the data
        #                 aux_priors_add[:, 0] = priors_remove[::num_try]

        #                 initial_log_proposal_pdf = self.gpu_priors["gb"].logpdf(new_coords_group_split)
        #                 # remove logpdf from fchange for now
        #                 initial_log_proposal_pdf -= self.gpu_priors["gb"].priors_in[1].logpdf(new_coords_group_split[:, 1])

        #                 aux_log_proposal_pdf[:, 0] = initial_log_proposal_pdf

        #                 aux_logP = band_inv_temps * aux_ll_diff + aux_priors_add

        #                 aux_log_importane_weights, aux_log_sum_weights, _ = get_mt_computations(aux_logP, aux_log_proposal_pdf, symmetric=False, xp=self.xp)

        #                 aux_logP_out = aux_logP[:, 0]
        #                 aux_log_proposal_pdf_out = aux_log_proposal_pdf[:, 0]

        #             factors = ((aux_logP_out - aux_log_sum_weights)- aux_log_proposal_pdf_out + aux_log_proposal_pdf_out) - ((logP_out - log_sum_weights) - log_proposal_pdf_out + log_proposal_pdf_out)

        #             lnpdiff = factors + logP_out - aux_logP_out

        #             keep = lnpdiff > self.xp.asarray(self.xp.log(self.xp.random.rand(*logP_out.shape)))

        #             coords_remove_keep = new_coords_group_split[keep]
        #             coords_add_keep = coords_add_out[keep]
        #             temp_inds_add_keep = temp_inds_add_group_split[keep]
        #             walker_inds_add_keep = walker_inds_add_group_split[keep]
        #             leaf_inds_add_keep = leaf_inds_add_group_split[keep]
        #             band_inds_add_keep = band_inds_add_group_split[keep]
        #             N_vals_add_keep = N_vals_add_group_split[keep]
        #             ll_diff_keep = ll_diff_out[keep]
        #             priors_add_keep = priors_add_out[keep]
        #             priors_remove_keep = priors_remove[::num_try][keep]

        #             # adjust everything
        #             ll_band_diff = xp.zeros((ntemps, nwalkers, len(self.band_edges) - 1))
        #             ll_band_diff[temp_inds_add_keep, walker_inds_add_keep, band_inds_add_keep] = ll_diff_keep
        #             lp_band_diff = xp.zeros((ntemps, nwalkers, len(self.band_edges) - 1))
        #             lp_band_diff[temp_inds_add_keep, walker_inds_add_keep, band_inds_add_keep] = priors_add_keep - priors_remove_keep

        #             new_state.branches["gb"].coords[temp_inds_add_keep.get(), walker_inds_add_keep.get(), band_inds_add_keep.get()] = coords_add_keep.get()
        #             new_state.log_like += ll_band_diff.sum(axis=-1).get()
        #             new_state.log_prior += ll_band_diff.sum(axis=-1).get()

        #             waveform_kwargs_tmp = self.waveform_kwargs.copy()
        #             waveform_kwargs_tmp.pop("N")
        #             coords_remove_keep_in = self.parameter_transforms.both_transforms(coords_remove_keep, xp=self.xp)
        #             coords_add_keep_in = self.parameter_transforms.both_transforms(coords_add_keep, xp=self.xp)
                   
        #             coords_in = xp.concatenate([coords_remove_keep_in, coords_add_keep_in], axis=0)
        #             factors = xp.concatenate([+xp.ones(coords_remove_keep_in.shape[0]), -xp.ones(coords_remove_keep_in.shape[0])])
        #             data_index_tmp = (temp_inds_add_keep * nwalkers + walker_inds_add_keep).astype(xp.int32)
        #             data_index_in = xp.concatenate([data_index_tmp, data_index_tmp], dtype=xp.int32)
        #             N_vals_in = xp.concatenate([N_vals_add_keep, N_vals_add_keep])
        #             self.gb.generate_global_template(
        #                 coords_in,
        #                 data_index_in,
        #                 self.mgh.data_list,
        #                 N=N_vals_in,
        #                 data_length=self.data_length,
        #                 data_splits=self.mgh.gpu_splits,
        #                 factors=factors,
        #                 **waveform_kwargs_tmp
        #             )
        #             self.xp.cuda.runtime.deviceSynchronize()

        #     ll_after = (
        #         self.mgh.get_ll(include_psd_info=True)
        #         .flatten()[new_state.supplimental[:]["overall_inds"]]
        #         .reshape(ntemps, nwalkers)
        #     )
        #     # print(np.abs(new_state.log_like - ll_after).max())
        #     store_max_diff = np.abs(new_state.log_like - ll_after).max()
        #     breakpoint()
        #     self.mempool.free_all_blocks()

                # self.mgh.restore_base_injections()

                # for name in new_state.branches.keys():
                #     if name not in ["gb", "gb"]:
                #         continue
                #     new_state_branch = new_state.branches[name]
                #     coords_here = new_state_branch.coords[new_state_branch.inds]
                #     ntemps, nwalkers, nleaves_max_here, ndim = new_state_branch.shape
                #     try:
                #         group_index = self.xp.asarray(
                #             self.mgh.get_mapped_indices(
                #                 np.repeat(
                #                     np.arange(ntemps * nwalkers).reshape(
                #                         ntemps, nwalkers, 1
                #                     ),
                #                     nleaves_max,
                #                     axis=-1,
                #                 )[new_state_branch.inds]
                #             ).astype(self.xp.int32)
                #         )
                #     except IndexError:
                #         breakpoint()
                #     coords_here_in = self.parameter_transforms.both_transforms(
                #         coords_here, xp=np
                #     )

                #     waveform_kwargs_fill = self.waveform_kwargs.copy()
                #     waveform_kwargs_fill["start_freq_ind"] = self.start_freq_ind

                #     if "N" in waveform_kwargs_fill:
                #         waveform_kwargs_fill.pop("N")

                #     self.mgh.multiply_data(-1.0)
                #     self.gb.generate_global_template(
                #         coords_here_in,
                #         group_index,
                #         self.mgh.data_list,
                #         data_length=self.data_length,
                #         data_splits=self.mgh.gpu_splits,
                #         **waveform_kwargs_fill
                #     )
                #     self.xp.cuda.runtime.deviceSynchronize()
                #     self.mgh.multiply_data(-1.0)





from multiprocessing.sharedctypes import Value
import os

import numpy as np

try:
    import cupy as xp

except (ImportError, ModuleNotFoundError) as e:
    pass

from eryn.state import State, BranchSupplimental
from eryn.utils.utility import groups_from_inds


class DetermineGBGroups:
    def __init__(self, gb_wave_generator, transform_fn=None, waveform_kwargs={}):
        self.gb_wave_generator = gb_wave_generator
        self.xp = self.gb_wave_generator.xp
        self.transform_fn = transform_fn
        self.waveform_kwargs = waveform_kwargs
            
    def __call__(self, last_sample, name_here, check_temp=0, input_groups=None, input_groups_inds=None, fix_group_count=False, mismatch_lim=0.2, double_check_lim=0.2, start_term="random", waveform_kwargs={}, index_within_group="random"):
        # TODO: mess with mismatch lim setting
        # TODO: some time of mismatch annealing may be useful
        if isinstance(last_sample, State):
            state = last_sample
            coords = state.branches_coords[name_here][check_temp]
            inds = state.branches_inds[name_here][check_temp]
        elif isinstance(last_sample, dict):
            coords = last_sample[name_here][check_temp]["coords"]
            inds = last_sample[name_here][check_temp]["inds"]

        waveform_kwargs = {**self.waveform_kwargs, **waveform_kwargs}
        
        # get coordinates and inds of the temperature you are considering.
        
        nwalkers, nleaves_max, ndim = coords.shape
        if input_groups is None:

            # figure our which walker to start with 
            if start_term == "max":
                start_walker_ind = inds[check_temp].sum(axis=-1).argmax()
            elif start_term == "first":
                start_walker_ind = 0
            elif start_term == "random":
                start_walker_ind = np.random.randint(0, nwalkers)
            else:
                raise ValueError("start_term must be 'max', 'first', or 'random'.")

            # get all the good leaves in this walker
            inds_good = np.where(inds[start_walker_ind])[0]
            groups = []
            groups_inds = []
            
            # set up this information to load the information into the group lists
            for leaf_i, leaf in enumerate(inds_good):
                groups.append([])
                groups_inds.append([])
                groups[leaf_i].append(coords[start_walker_ind, leaf].copy())
                groups_inds[leaf_i].append([start_walker_ind, leaf])
        else:
            # allows us to check groups based on groups we already have
            groups = input_groups
            groups_inds = input_groups_inds

        if len(groups) == 0:
            return [], [], []
        for w in range(coords.shape[0]):

            # we have already loaded this group
            if input_groups is None and w == start_walker_ind:
                continue

            # walker has no binaries 
            if not np.any(inds[w]):
                continue

            # coords in this walker
            coords_here = coords[w][inds[w]]
            inds_for_group_stuff = np.arange(len(inds[w]))[inds[w]]
            nleaves, ndim = coords_here.shape

            
            params_for_test = []
            for group in groups:
                group_params = np.asarray(group)

                if index_within_group == "first":
                    test_walker_ind = 0
                elif index_within_group == "random":
                    test_walker_ind = np.random.randint(0, group_params.shape[0])
                else:
                    raise ValueError("start_term must be 'max', 'first', or 'random'.")

                params_for_test.append(group_params[test_walker_ind])
            params_for_test = np.asarray(params_for_test)
            
            # transform coords
            if self.transform_fn is not None:
                params_for_test_in = self.transform_fn[name_here].both_transforms(params_for_test, return_transpose=False)
                coords_here_in = self.transform_fn[name_here].both_transforms(coords_here, return_transpose=False)
            
            else:
                params_for_test_in = params_for_test.copy()
                coords_here_in = coords_here.copy()

            inds_tmp_test = np.arange(len(params_for_test_in))
            inds_tmp_here = np.arange(len(coords_here_in))
            inds_tmp_test, inds_tmp_here = [tmp.ravel() for tmp in np.meshgrid(inds_tmp_test, inds_tmp_here)]

            params_for_test_in_full = params_for_test_in[inds_tmp_test]
            coords_here_in_full = coords_here_in[inds_tmp_here]
            # build the waveforms at the same time

            df = 1. / waveform_kwargs["T"]
            max_f = 1. / 2 * 1/waveform_kwargs["dt"]
            frqs = self.xp.arange(0.0, max_f, df)
            data_minus_template = self.xp.asarray([
                self.xp.ones_like(frqs, dtype=complex),
                self.xp.ones_like(frqs, dtype=complex)
            ])[None, :, :]
            psd = self.xp.asarray([
                self.xp.ones_like(frqs, dtype=np.float64),
                self.xp.ones_like(frqs, dtype=np.float64)
            ])

            waveform_kwargs_fill = waveform_kwargs.copy()
            waveform_kwargs_fill.pop("start_freq_ind")
 
            # TODO: could use real data and get observed snr for each if needed
            check = self.gb_wave_generator.swap_likelihood_difference(
                params_for_test_in_full,
                coords_here_in_full,
                data_minus_template,
                psd,
                start_freq_ind=0,
                data_index=None,
                noise_index=None,
                **waveform_kwargs_fill,
            )

            numerator = self.gb_wave_generator.add_remove
            norm_here = self.gb_wave_generator.add_add
            norm_for_test = self.gb_wave_generator.remove_remove

            normalized_autocorr = numerator / np.sqrt(norm_here * norm_for_test)
            normalized_against_test = numerator / norm_for_test

            normalized_autocorr = normalized_autocorr.reshape(coords_here_in.shape[0], params_for_test_in.shape[0]).real
            normalized_against_test = normalized_against_test.reshape(coords_here_in.shape[0], params_for_test_in.shape[0]).real
            
            # TODO: do based on Likelihood? make sure on same posterior
            # TODO: add check based on amplitude
            test1 = np.abs(1.0 - normalized_autocorr.real)  # (numerator / norm_for_test[None, :]).real)
            best = test1.argmin(axis=1)
            try:
                best = best.get()
            except AttributeError:
                pass
            best_mismatch = test1[(np.arange(test1.shape[0]), best)]
            check_normalized_against_test = np.abs(1.0 - normalized_against_test[(np.arange(test1.shape[0]), best)])


            f0_here = coords_here[:, 1]
            f0_test = params_for_test[best, 1]

            for leaf in range(nleaves):
                if best_mismatch[leaf] < mismatch_lim and check_normalized_against_test[leaf] < double_check_lim:
                    groups[best[leaf]].append(coords_here[leaf].copy())
                    groups_inds[best[leaf]].append([w, inds_for_group_stuff[leaf]])

                elif not fix_group_count:
                    # this only works for high snr limit
                    groups.append([coords_here[leaf]].copy())
                    groups_inds.append([[w, inds_for_group_stuff[leaf]]])

        group_lens = [len(group) for group in groups]

        return groups, groups_inds, group_lens


class GetLastGBState:
    def __init__(self, gb_wave_generator, transform_fn=None, waveform_kwargs={}):
        self.gb_wave_generator = gb_wave_generator
        self.xp = self.gb_wave_generator.xp
        self.transform_fn = transform_fn
        self.waveform_kwargs = waveform_kwargs

    def __call__(self, mgh, reader, df, supps_base_shape, fix_temp_initial_ind:int=None, fix_temp_inds:list=None, nleaves_max_in=None, waveform_kwargs={}):

        xp.cuda.runtime.setDevice(mgh.gpus[0])

        if fix_temp_initial_ind is not None or fix_temp_inds is not None:
            if fix_temp_initial_ind is None or fix_temp_inds is None:
                raise ValueError("If giving fix_temp_initial_ind or fix_temp_inds, must give both.")

        state = reader.get_last_sample()

        waveform_kwargs = {**self.waveform_kwargs, **waveform_kwargs}
        if "start_freq_ind" not in waveform_kwargs:
            raise ValueError("In get_last_gb_state, waveform_kwargs must include 'start_freq_ind'.")

        #check = reader.get_last_sample()
        ntemps, nwalkers, nleaves_max_old, ndim = state.branches["gb"].shape
        
        #out = get_groups_for_remixing(check, check_temp=0, input_groups=None, input_groups_inds=None, fix_group_count=False, name_here="gb")

        #lengths = []
        #for group in out[0]:
        #    lengths.append(len(group))
        #breakpoint()
        try:
            if fix_temp_initial_ind is not None: 
                for i in fix_temp_inds:
                    if i < fix_temp_initial_ind:
                        raise ValueError("If providing fix_temp_initial_ind and fix_temp_inds, all values in fix_temp_inds must be greater than fix_temp_initial_ind.")

                    state.log_like[i] = state.log_like[fix_temp_initial_ind]
                    state.log_prior[i] = state.log_prior[fix_temp_initial_ind]
                    state.branches_coords["gb"][i] = state.branches_coords["gb"][fix_temp_initial_ind]
                    state.branches_coords["gb"][i] = state.branches_coords["gb"][fix_temp_initial_ind]
                    state.branches_inds["gb"][i] = state.branches_inds["gb"][fix_temp_initial_ind]
                    state.branches_inds["gb"][i] = state.branches_inds["gb"][fix_temp_initial_ind]

            ntemps, nwalkers, nleaves_max_old, ndim = state.branches["gb"].shape
            if nleaves_max_in is None:
                nleaves_max = nleaves_max_old
            else:
                nleaves_max = nleaves_max_in
            if nleaves_max_old <= nleaves_max:
                coords_tmp = np.zeros((ntemps, nwalkers, nleaves_max, ndim))
                coords_tmp[:, :, :nleaves_max_old, :] = state.branches["gb"].coords

                inds_tmp = np.zeros((ntemps, nwalkers, nleaves_max), dtype=bool)
                inds_tmp[:, :, :nleaves_max_old] = state.branches["gb"].inds
                state.branches["gb"].coords = coords_tmp
                state.branches["gb"].inds = inds_tmp
                state.branches["gb"].nleaves_max = nleaves_max
                state.branches["gb"].shape = (ntemps, nwalkers, nleaves_max, ndim)
                
            else:
                raise ValueError("new nleaves_max is less than nleaves_max_old.")

            # add "gb" if there are any
            data_index_in = groups_from_inds({"gb": state.branches_inds["gb"]})["gb"]

            data_index = xp.asarray(mgh.get_mapped_indices(data_index_in)).astype(xp.int32)

            params_add_in = self.transform_fn["gb"].both_transforms(state.branches_coords["gb"][state.branches_inds["gb"]])
            
            # batch_size is ignored if waveform_kwargs["use_c_implementation"] is True
            #  -1 is to do -(-d + h) = d - h  
            mgh.multiply_data(-1.)
            self.gb_wave_generator.generate_global_template(params_add_in, data_index, mgh.data_list, data_length=mgh.data_length, data_splits=mgh.gpu_splits, batch_size=1000, **waveform_kwargs)
            mgh.multiply_data(-1.)
            

        except KeyError:
            # no "gb"
            pass

        data_index_in = groups_from_inds({"gb": state.branches_inds["gb"]})["gb"]
        data_index = xp.asarray(mgh.get_mapped_indices(data_index_in)).astype(xp.int32)

        params_add_in = self.transform_fn["gb"].both_transforms(state.branches_coords["gb"][state.branches_inds["gb"]])
        
        #  -1 is to do -(-d + h) = d - h  
        mgh.multiply_data(-1.)
        self.gb_wave_generator.generate_global_template(params_add_in, data_index, mgh.data_list, data_length=mgh.data_length, data_splits=mgh.gpu_splits, batch_size=1000, **waveform_kwargs)
        mgh.multiply_data(-1.)

        self.gb_wave_generator.d_d = np.asarray(mgh.get_inner_product(use_cpu=True))
        
        state.log_like = -1/2 * self.gb_wave_generator.d_d.real.reshape(ntemps, nwalkers)

        temp_inds = mgh.temp_indices.copy()
        walker_inds = mgh.walker_indices.copy()
        overall_inds = mgh.overall_indices.copy()
            
        supps = BranchSupplimental({ "temp_inds": temp_inds, "walker_inds": walker_inds, "overall_inds": overall_inds,}, obj_contained_shape=supps_base_shape, copy=True)
        state.supplimental = supps

        return state


class HeterodynedUpdate:
    def __init__(self, update_kwargs, set_d_d_zero=False):
        self.update_kwargs = update_kwargs
        self.set_d_d_zero = set_d_d_zero

    def __call__(self, it, sample_state, sampler, **kwargs):

        samples = sample_state.branches_coords["mbh"].reshape(-1, sampler.ndims[0])
        lp_max = sample_state.log_like.argmax()
        best = samples[lp_max]

        lp = sample_state.log_like.flatten()
        sorted = np.argsort(lp)
        inds_best = sorted[-1000:]
        inds_worst = sorted[:1000]

        best_full = sampler.log_like_fn.f.parameter_transforms["mbh"].both_transforms(
            best, copy=True
        )

        sampler.log_like_fn.f.template_model.init_heterodyne_info(
            best_full, **self.update_kwargs
        )

        if self.set_d_d_zero:
            sampler.log_like_fn.f.template_model.reference_d_d = 0.0

        # TODO: make this a general update function in Eryn (?)
        # samples[inds_worst] = samples[inds_best].copy()
        samples = samples.reshape(sampler.ntemps, sampler.nwalkers, 1, sampler.ndims[0])
        logp = sampler.compute_log_prior({"mbh": samples})
        logL, blobs = sampler.compute_log_like({"mbh": samples}, logp=logp)

        sample_state.branches["mbh"].coords = samples
        sample_state.log_like = logL
        sample_state.blobs = blobs

        # sampler.backend.save_step(sample_state, np.full_like(lp, True))



import numpy as np
from scipy import stats

from ..utils.constants import *
from ..sensitivity import get_sensitivity
from eryn.moves.multipletry import logsumexp

from typing import Union, Optional, Tuple, List

import sys
sys.path.append("/data/mkatz/LISAanalysistools/lisaflow/flow/experiments/rvs/gf_search/")
# from galaxy_ffdot import GalaxyFFdot
# from galaxy import Galaxy

try:
    import cupy as cp

except (ModuleNotFoundError, ImportError) as e:
    pass

class AmplitudeFrequencySNRPrior:
    def __init__(self, rho_star, frequency_prior, L, Tobs, use_cupy=False, **noise_kwargs):
        self.rho_star = rho_star
        self.frequency_prior = frequency_prior

        self.transform = AmplitudeFromSNR(L, Tobs, use_cupy=use_cupy, **noise_kwargs)
        self.snr_prior = SNRPrior(rho_star, use_cupy=use_cupy)
        
        # must be after transform and snr_prior due to setter
        self.use_cupy = use_cupy
        
    @property
    def use_cupy(self):
        return self._use_cupy

    @use_cupy.setter
    def use_cupy(self, use_cupy):
        self._use_cupy = use_cupy
        self.transform.use_cupy = use_cupy
        self.snr_prior.use_cupy = use_cupy
        self.frequency_prior.use_cupy = use_cupy

    def pdf(self, *args, **noise_kwargs):
        return np.exp(self.logpdf(*args, **noise_kwargs))

    def logpdf(self, amp, f0_ms, **noise_kwargs):

        xp = np if not self.use_cupy else cp

        f0 = f0_ms / 1e3
        rho, f0 = self.transform.forward(amp, f0, **noise_kwargs)

        rho_pdf = self.snr_prior.pdf(rho)

        Jac = xp.abs(rho / amp)

        logpdf_amp = np.log(np.abs(Jac * rho_pdf))
        logpdf_f = self.frequency_prior.logpdf(f0_ms)

        return logpdf_amp + logpdf_f

    def rvs(self, size=1, f0_input=None, **noise_kwargs):
        if isinstance(size, int):
            size = (size,)

        xp = np if not self.use_cupy else cp

        if f0_input is None:
            f0_ms = self.frequency_prior.rvs(size=size)
        else:
            f0_ms = f0_input
            assert f0_input.shape[:-1] == size
        
        f0 = f0_ms / 1e3

        rho = self.snr_prior.rvs(size=size)

        amp, _ = self.transform(rho, f0, **noise_kwargs)

        return (amp, f0_ms)






class SNRPrior:
    def __init__(self, rho_star, use_cupy=False):
        self.rho_star = rho_star
        self.use_cupy = use_cupy

    @property
    def use_cupy(self):
        return self._use_cupy

    @use_cupy.setter
    def use_cupy(self, use_cupy):
        self._use_cupy = use_cupy

    def pdf(self, rho):
        
        xp = np if not self.use_cupy else cp

        p = xp.zeros_like(rho)
        good = rho > 0.0
        p[good] = 3 * rho[good] / (4 * self.rho_star ** 2 * (1 + rho[good] / (4 * self.rho_star)) ** 5)
        return p

    def logpdf(self, rho):
        xp = np if not self.use_cupy else cp
        return xp.log(self.pdf(rho))

    def cdf(self, rho):
        xp = np if not self.use_cupy else cp
        c = xp.zeros_like(rho)
        good = rho > 0.0
        c[good] = 768 * self.rho_star ** 3 * (1 / (768. * self.rho_star ** 3) - (rho[good] + self.rho_star)/(3. * (rho[good] + 4 * self.rho_star) ** 4))
        return c

    def rvs(self, size=1):
        if isinstance(size, int):
            size = (size,)

        xp = np if not self.use_cupy else cp

        u = xp.random.rand(*size)

        rho = (-4*self.rho_star + xp.sqrt(-32*self.rho_star**2 - (32*(-self.rho_star**2 + u*self.rho_star**2))/(1 - u) + 
      (3072*2**0.3333333333333333*xp.cbrt(-1 + 3*u - 3*u**2 + u**3)*
         (self.rho_star**4 - u*self.rho_star**4))/
       ((-1 + u)**2*xp.cbrt(-1769472*self.rho_star**6 + 1769472*u*self.rho_star**6 - 
            xp.sqrt(3131031158784*u*self.rho_star**12 - 6262062317568*u**2*self.rho_star**12 + 
              3131031158784*u**3*self.rho_star**12))) + 
      xp.cbrt(-1769472*self.rho_star**6 + 1769472*u*self.rho_star**6 - 
          xp.sqrt(3131031158784*u*self.rho_star**12 - 6262062317568*u**2*self.rho_star**12 + 
            3131031158784*u**3*self.rho_star**12))/
       (3.*2**0.3333333333333333*xp.cbrt(-1 + 3*u - 3*u**2 + u**3)))/2.
     + xp.sqrt(32*self.rho_star**2 + (32*(-self.rho_star**2 + u*self.rho_star**2))/(1 - u) - 
      (3072*2**0.3333333333333333*xp.cbrt(-1 + 3*u - 3*u**2 + u**3)*
         (self.rho_star**4 - u*self.rho_star**4))/
       ((-1 + u)**2*xp.cbrt(-1769472*self.rho_star**6 + 1769472*u*self.rho_star**6 - 
            xp.sqrt(3131031158784*u*self.rho_star**12 - 6262062317568*u**2*self.rho_star**12 + 
              3131031158784*u**3*self.rho_star**12))) - 
      xp.cbrt(-1769472*self.rho_star**6 + 1769472*u*self.rho_star**6 - 
          xp.sqrt(3131031158784*u*self.rho_star**12 - 6262062317568*u**2*self.rho_star**12 + 
            3131031158784*u**3*self.rho_star**12))/
       (3.*2**0.3333333333333333*xp.cbrt(-1 + 3*u - 3*u**2 + u**3)) + 
      (2048*self.rho_star**3 - (2048*u*self.rho_star**3)/(-1 + u))/
       (4.*xp.sqrt(-32*self.rho_star**2 - (32*(-self.rho_star**2 + u*self.rho_star**2))/(1 - u) + 
           (3072*2**0.3333333333333333*
              xp.cbrt(-1 + 3*u - 3*u**2 + u**3)*(self.rho_star**4 - u*self.rho_star**4)
              )/
            ((-1 + u)**2*xp.cbrt(-1769472*self.rho_star**6 + 1769472*u*self.rho_star**6 - 
                 xp.sqrt(3131031158784*u*self.rho_star**12 - 6262062317568*u**2*self.rho_star**12 + 
                   3131031158784*u**3*self.rho_star**12))) + 
           xp.cbrt(-1769472*self.rho_star**6 + 1769472*u*self.rho_star**6 - 
               xp.sqrt(3131031158784*u*self.rho_star**12 - 6262062317568*u**2*self.rho_star**12 + 
                 3131031158784*u**3*self.rho_star**12))/
            (3.*2**0.3333333333333333*
              xp.cbrt(-1 + 3*u - 3*u**2 + u**3)))))/2.)

        return rho


class AmplitudeFromSNR:
    def __init__(self, L, Tobs, fd=None, use_cupy=False, **noise_kwargs):
        self.f_star = 1 / (2. * np.pi * L) * C_SI
        self.Tobs = Tobs
        self.noise_kwargs = noise_kwargs
        
        xp = np if not use_cupy else cp
        if fd is not None:
            self.fd = xp.asarray(fd)
        else:
            self.fd = fd

        # got to be after fd
        self.use_cupy = use_cupy

    @property
    def use_cupy(self):
        return self._use_cupy

    @use_cupy.setter
    def use_cupy(self, use_cupy):
        self._use_cupy = use_cupy
        if use_cupy and not isinstance(self.fd, cp.ndarray):
            self.fd = cp.asarray(self.fd)
        elif not use_cupy and isinstance(self.fd, cp.ndarray):
            self.fd = self.fd.get()

    def interp_psd(self, f0, psds, walker_inds=None):
        assert self.fd is not None
        xp = np if not self.use_cupy else cp
        psds = xp.atleast_2d(psds)
        
        if xp == cp and not isinstance(self.fd, cp.ndarray):
            self.fd = xp.asarray(self.fd)
        try:
            inds_fd = xp.searchsorted(self.fd, f0, side="right") - 1
        except:
            breakpoint()
        if walker_inds is None:
            walker_inds = xp.zeros_like(f0, dtype=int)

        new_psds = (psds[(walker_inds, inds_fd + 1)] - psds[(walker_inds, inds_fd)]) / (self.fd[inds_fd + 1] - self.fd[inds_fd]) * (f0 - self.fd[inds_fd]) + psds[(walker_inds, inds_fd)]
        return new_psds

    def __call__(self, rho, f0, **noise_kwargs):

        xp = np if not self.use_cupy else cp

        if noise_kwargs == {}:
            noise_kwargs = self.noise_kwargs

        Sn_f = self.get_Sn_f(f0, **noise_kwargs)

        factor = 1./2. * np.sqrt((self.Tobs * np.sin(f0 / self.f_star) ** 2) / Sn_f)
        amp = rho / factor
        return (amp, f0)

    def get_Sn_f(self, f0, psds=None, walker_inds=None, Sn_f=None, **noise_kwargs):
        if Sn_f is not None:
            assert len(f0) == len(Sn_f)
            assert isinstance(f0, type(Sn_f))

        elif psds is not None:
            Sn_f = self.interp_psd(f0, psds, walker_inds=walker_inds)
        else:
            Sn_f = get_sensitivity(f0, **noise_kwargs)

        return Sn_f

    def forward(self, amp, f0, **noise_kwargs):

        if noise_kwargs == {}:
            noise_kwargs = self.noise_kwargs

        Sn_f = self.get_Sn_f(f0, **noise_kwargs)

        factor = 1./2. * np.sqrt((self.Tobs * np.sin(f0 / self.f_star) ** 2) / Sn_f)
        rho = amp * factor
        return (rho, f0)


class GBPriorWrap:
    def __init__(self, ndim, full_prior_container, gen_frequency_alone=False):
        self.base_prior = full_prior_container
        self.use_cupy = full_prior_container.use_cupy
        self.ndim = ndim
        self.gen_frequency_alone = gen_frequency_alone

        if gen_frequency_alone:
            self.keys_sep = [1, 2, 3, 4, 5, 6, 7]
        else:
            self.keys_sep = [2, 3, 4, 5, 6, 7]

    @property
    def priors_in(self):
        return self.base_prior.priors_in

    def logpdf(self, x, **noise_kwargs):
        xp = np if not self.use_cupy else cp
        assert x.shape[1] == self.ndim and x.ndim == 2

        logpdf_everything_else = self.base_prior.logpdf(x, keys=self.keys_sep)

        f0 = xp.asarray(x[:, 1])
        amp = xp.asarray(x[:, 0])
        logpdf_A_f = self.base_prior.priors_in[(0, 1)].logpdf(amp, f0, **noise_kwargs)

        return logpdf_A_f + logpdf_everything_else

    def rvs(self, size=1, ignore_amp=False, **kwargs):
        xp = np if not self.use_cupy else cp
        if isinstance(size, int):
            size = (size,)
        
        arr = xp.zeros(size + (self.ndim,)).reshape(-1, self.ndim)

        diff = self.ndim - len(self.keys_sep)
        assert diff >= 0

        arr[:, :] = self.base_prior.rvs(size, keys=self.keys_sep).reshape(-1, self.ndim)

        if not ignore_amp:
            f0_input = arr[:, 1] if self.gen_frequency_alone else None
            arr[:, :diff] = xp.asarray(self.base_prior.priors_in[(0, 1)].rvs(size, f0_input=f0_input, **kwargs)).reshape(diff, -1).T

        arr = arr.reshape(size + (self.ndim,))
        return arr


class FullGaussianMixtureModel:
    def __init__(self, gb, weights, means, covs, invcovs, dets, mins, maxs, limit=10.0, use_cupy=False):
        
        self.use_cupy = use_cupy
        if use_cupy:
            xp = cp
        else:
            xp = np

        self.gb = gb

        indexing = []
        for i, weight in enumerate(weights):
            index_base = np.full_like(weight, i, dtype=int)
            indexing.append(index_base)
        
        self.indexing = xp.asarray(np.concatenate(indexing))
        # invidivual weights / total number of components to uniformly choose from them
        self.weights = xp.asarray(np.concatenate(weights, axis=0) * 1 / len(weights))

        assert xp.allclose(self.weights.sum(), 1.0)
        self.means = xp.asarray(np.concatenate(means, axis=0))
        self.covs = xp.asarray(np.concatenate(covs, axis=0))
        self.invcovs = xp.asarray(np.concatenate(invcovs, axis=0))
        self.dets = xp.asarray(np.concatenate(dets, axis=0))
        self.ndim = self.means.shape[1]

        self.mins = xp.asarray(np.vstack(mins))
        self.maxs = xp.asarray(np.vstack(maxs))

        self.mins_in_pdf = self.mins[self.indexing].T.flatten().copy()
        self.maxs_in_pdf = self.maxs[self.indexing].T.flatten().copy()
        self.means_in_pdf = self.means.T.flatten().copy()
        self.invcovs_in_pdf = self.invcovs.transpose(1, 2, 0).flatten().copy()

        self.cumulative_weights = xp.concatenate([xp.array([0.0]), xp.cumsum(self.weights)])

        self.min_limit_f = self.map_back_frequency(-1. * limit, self.mins[self.indexing, 1], self.maxs[self.indexing, 1]) 
        self.max_limit_f = self.map_back_frequency(+1. * limit, self.mins[self.indexing, 1], self.maxs[self.indexing, 1]) 

        # compute the jacobian
        self.log_det_J = (self.ndim * np.log(2) - xp.sum(xp.log(self.maxs - self.mins), axis=-1))[self.indexing].copy()

        """self.inds_sort_min_limit_f = xp.argsort(self.min_limit_f)
        self.inds_sort_max_limit_f = xp.argsort(self.max_limit_f)
        self.sorted_min_limit_f = self.min_limit_f[self.inds_sort_min_limit_f]
        self.sorted_max_limit_f = self.max_limit_f[self.inds_sort_max_limit_f]
        """
    def logpdf(self, x):

        if self.use_cupy:
            xp = cp
        else:
            xp = np

        assert len(x.shape) == 2
        assert x.shape[1] == self.ndim

        k = self.ndim

        inds_sort = xp.argsort(x[:, 1])
        f_sort = x[:, 1][inds_sort]
        points_sorted = x[inds_sort]

        ind_min_limit = xp.searchsorted(f_sort, self.min_limit_f, side="left")
        ind_max_limit = xp.searchsorted(f_sort, self.max_limit_f, side="right")

        diff = (ind_max_limit - ind_min_limit)
        cs = xp.concatenate([xp.array([0]), xp.cumsum(diff)])
        tmp = xp.arange(cs[-1])
        keep_component_map = xp.searchsorted(cs, tmp, side="right") - 1
        keep_point_map = tmp - cs[keep_component_map] + ind_min_limit[keep_component_map]
        max_components = diff.max().item()
        
        int_check = int(1e6)
        assert int_check > self.min_limit_f.shape[0]
        special_point_component_map = int_check * keep_point_map + keep_component_map

        sorted_special = xp.sort(special_point_component_map)

        points_keep_in = (sorted_special / float(int_check)).astype(int)
        components_keep_in = sorted_special - points_keep_in * int_check

        unique_points, unique_starts = xp.unique(points_keep_in, return_index=True)
        start_index_in_pdf = xp.concatenate([unique_starts, xp.array([len(points_keep_in)])]).astype(xp.int32)
        assert xp.all(xp.diff(unique_starts) > 0)
        
        points_sorted_in = points_sorted[unique_points]

        logpdf_out_tmp = xp.zeros(points_sorted_in.shape[0])

        self.gb.compute_logpdf(logpdf_out_tmp, components_keep_in.astype(xp.int32), points_sorted_in,
                    self.weights, self.mins_in_pdf, self.maxs_in_pdf, self.means_in_pdf, self.invcovs_in_pdf, self.dets, self.log_det_J, 
                    points_sorted_in.shape[0], start_index_in_pdf, self.weights.shape[0], x.shape[1])

        # need to reverse the sort
        logpdf_out = xp.full(x.shape[0], -xp.inf)
        logpdf_out[xp.sort(inds_sort[unique_points])] = logpdf_out_tmp[xp.argsort(inds_sort[unique_points])]
        return logpdf_out
        """# breakpoint()

        # map to reduced domain
        x_mapped = (self.map_input(points_sorted[:, None, :], self.mins[None, :, :], self.maxs[None, :, :]))[:, self.indexing]
        
        diff = x_mapped - self.means[None, :, :]
        log_main_part = -1./2. * xp.einsum("...k,...k", diff, xp.einsum("...jk,...k->...j", self.invcovs, diff))
        log_norm_factor = (k / 2) * xp.log(2 * np.pi) + (1 / 2) * xp.log(self.dets)
        log_weighted_pdf = (xp.log(self.weights) + log_norm_factor)[None, :] + log_main_part
        
        logpdf_full_dist_tmp = logsumexp(log_weighted_pdf, axis=-1, xp=xp)
        logpdf_full_dist = logpdf_full_dist_tmp[xp.argsort(inds_sort)]

        breakpoint()
        assert xp.allclose(logpdf_full_dist, logpdf_out)

        return logpdf_full_dist"""

    def map_input(self, x, mins, maxs):
        return ((x - mins) / (maxs - mins)) * 2. - 1.

    def map_back_frequency(self, x, mins, maxs):
        return (x + 1.) * 1. / 2. * (maxs - mins) + mins
        
    def rvs(self, size=(1,)):

        if isinstance(size, int):
            size = (size,)

        if self.use_cupy:
            xp = cp
        else:
            xp = np

        # choose which component
        draw = xp.random.rand(*size)
        component = (xp.searchsorted(self.cumulative_weights, draw.flatten(), side="right") - 1).reshape(draw.shape)

        mean_here = self.means[component]
        cov_here = self.covs[component]

        new_points = mean_here + xp.einsum("...kj,...j->...k", cov_here, np.random.randn(*(component.shape + (self.ndim,))))

        index_here = self.indexing[component]
        mins_here = self.mins[index_here]
        maxs_here = self.maxs[index_here]
        new_points_mapped = self.map_back_frequency(new_points, mins_here, maxs_here)
        
        return new_points_mapped
        

# class FlowDist:
#     def __init__(self, config: dict, model: Union[Galaxy, GalaxyFFdot], fit: str, ndim: int):

#         self.dist = model(config)
#         self.dist.load_fit()

#         param_min, param_max = np.loadtxt(fit)
#         self.dist.set_min(param_min)
#         self.dist.set_max(param_max)
        
#         self.config = config
#         self.fit = fit
#         self.ndim = ndim

#     def rvs(self, size: Optional[Union[int, tuple]]=(1,)) -> cp.ndarray:
#         if isinstance(size, int):
#             size = (size,)

#         total_samp = int(np.prod(size))
#         samples = self.dist.sample(total_samp).reshape(size + (self.ndim,))
#         return samples 

#     def logpdf(self, x: cp.ndarray) -> cp.ndarray:
#         assert x.shape[-1] == self.ndim
#         log_prob = self.dist.log_prob(x.reshape(-1, self.ndim)).reshape(x.shape[:-1])
#         return log_prob

# class GalaxyFlowDist(FlowDist):
#     def __init__(self):
#         config = '/data/mkatz/LISAanalysistools/lisaflow/flow/experiments/configs/gbs/density_galaxy.yaml'
#         model = Galaxy
#         fit = '/data/mkatz/LISAanalysistools/lisaflow/flow/experiments/rvs/minmax_galaxy_sangria.txt'
#         ndim = 3
#         super().__init__(config, model, fit, ndim)

#     def logpdf(self, x: cp.ndarray) -> cp.ndarray:
#         # adjust amplitudes to exp
#         x[:, 0] = np.log(x[:, 0])
#         return super().logpdf(x)

#     def rvs(self, size: Optional[Union[int, tuple]]=(1,)) -> cp.ndarray:
#         if isinstance(size, int):
#             size = (size,)

#         samples = super().rvs(size=size)
#         samples = samples.reshape(-1, samples.shape[-1])
#         samples[:, 0] = np.exp(samples[:, 0])
#         samples = samples.reshape(size + (samples.shape[-1],))
#         return samples

# class FFdotFlowDist(FlowDist):
#     def __init__(self):
#         config = '/data/mkatz/LISAanalysistools/lisaflow/flow/experiments/configs/gbs/density_f.yaml'
#         model = GalaxyFFdot
#         fit = '/data/mkatz/LISAanalysistools/lisaflow/flow/experiments/rvs/minmax_ffdot_sangria.txt'
#         ndim = 2
#         super().__init__(config, model, fit, ndim)
        
    

    
        
        
            
            


import warnings
from abc import ABC
from typing import Any, Tuple, Optional, List

import math
import numpy as np
from numpy.typing import ArrayLike
from scipy import interpolate
import matplotlib.pyplot as plt

try:
    import cupy as cp

except (ModuleNotFoundError, ImportError):
    import numpy as cp

from . import detector as lisa_models
from .utils.utility import AET, get_array_module
from .utils.constants import *
from .stochastic import (
    StochasticContribution,
    FittedHyperbolicTangentGalacticForeground,
)
from .sensitivity import SensitivityMatrix


class DataResidualArray:
    pass


class DataResidualArray:
    """Container to hold sensitivity information.

    Args:
        f: Frequency array.
        sens_mat: Input sensitivity list. The shape of the nested lists should represent the shape of the
            desired matrix. Each entry in the list must be an array, :class:`Sensitivity`-derived object,
            or a string corresponding to the :class:`Sensitivity` object.
        **sens_kwargs: Keyword arguments to pass to :method:`Sensitivity.get_Sn`.

    """

    def __init__(
        self,
        data_res_in: List[np.ndarray] | np.ndarray | DataResidualArray,
        dt: Optional[float] = None,
        f_arr: Optional[np.ndarray] = None,
        df: Optional[float] = None,
        **kwargs: dict,
    ) -> None:
        if isinstance(data_res_in, DataResidualArray):
            for key, item in data_res_in.__dict__.items():
                setattr(self, key, item)

        else:
            self._check_inputs(dt=dt, f_arr=f_arr, df=df)
            self.data_res_arr = data_res_in
            self._store_time_and_frequency_information(dt=dt, f_arr=f_arr, df=df)

    @property
    def init_kwargs(self) -> dict:
        """Initial dt, df, f_arr"""
        return self._init_kwargs

    @init_kwargs.setter
    def init_kwargs(self, init_kwargs: dict) -> None:
        """Set initial kwargs."""
        self._init_kwargs = init_kwargs

    def _check_inputs(
        self,
        dt: Optional[float] = None,
        f_arr: Optional[np.ndarray] = None,
        df: Optional[float] = None,
    ):
        number_of_none = 0

        number_of_none += 1 if dt is None else 0
        number_of_none += 1 if f_arr is None else 0
        number_of_none += 1 if df is None else 0

        if number_of_none == 3:
            raise ValueError("Must provide either df, dt, or f_arr.")

        elif number_of_none == 1:
            raise ValueError(
                "Can only provide one of dt, f_arr, or df. Not more than one."
            )
        self.init_kwargs = dict(dt=dt, f_arr=f_arr, df=df)

    def _store_time_and_frequency_information(
        self,
        dt: Optional[float] = None,
        f_arr: Optional[np.ndarray] = None,
        df: Optional[float] = None,
    ):
        if dt is not None:
            self._dt = dt
            self._Tobs = self.data_length * dt
            self._df = 1 / self._Tobs
            self._fmax = 1 / (2 * dt)
            self._f_arr = np.fft.rfftfreq(self.data_length, dt)

            # transform data
            tmp = (
                get_array_module(self.data_res_arr).fft.rfft(self.data_res_arr, axis=-1)
                * self._dt
            )
            del self._data_res_arr
            self._data_res_arr = tmp
            self.data_length = self._data_res_arr.shape[-1]

        elif df is not None:
            self._df = df
            self._Tobs = 1 / self._df
            self._fmax = (self.data_length - 1) * df
            self._dt = 1 / (2 * self._fmax)
            self._f_arr = np.arange(0.0, self._fmax, self._df)

        elif f_arr is not None:
            self._f_arr = f_arr
            self._fmax = f_arr.max()
            # constant spacing
            if np.all(np.diff(f_arr) == np.diff(f_arr)[0]):
                self._df = np.diff(f_arr)[0].item()

                if f_arr[0] == 0.0:
                    # could be fft because of constant spacing and f_arr[0] == 0.0
                    self._Tobs = 1 / self._df
                    self._dt = 1 / (2 * self._fmax)

                else:
                    # cannot be fft basis
                    self._Tobs = None
                    self._dt = None

            else:
                self._df = None
                self._Tobs = None
                self._dt = None

        if len(self.f_arr) != self.data_length:
            raise ValueError(
                "Entered or determined f_arr does not have the same length as the data channel inputs."
            )

    @property
    def fmax(self):
        return self._fmax

    @property
    def f_arr(self):
        return self._f_arr

    @property
    def dt(self):
        if self._dt is None:
            raise ValueError("dt cannot be determined from this f_arr input.")

        return self._dt

    @property
    def Tobs(self):
        if self._Tobs is None:
            raise ValueError("Tobs cannot be determined from this f_arr input.")

        return self._Tobs

    @property
    def df(self):
        if self._df is None:
            raise ValueError("df cannot be determined from this f_arr input.")

        return self._df

    @property
    def frequency_arr(self) -> np.ndarray:
        return self._f_arr

    @property
    def data_res_arr(self) -> np.ndarray:
        return self._data_res_arr

    @data_res_arr.setter
    def data_res_arr(self, data_res_arr: List[np.ndarray] | np.ndarray) -> None:
        self._data_res_arr_input = data_res_arr

        if (
            isinstance(data_res_arr, np.ndarray) or isinstance(data_res_arr, cp.ndarray)
        ) and data_res_arr.ndim == 1:
            data_res_arr = [data_res_arr]

        elif (
            isinstance(data_res_arr, np.ndarray) or isinstance(data_res_arr, cp.ndarray)
        ) and data_res_arr.ndim == 2:
            data_res_arr = list(data_res_arr)

        new_out = np.full(len(data_res_arr), None, dtype=object)
        self.data_length = None
        for i in range(len(data_res_arr)):
            current_data = data_res_arr[i]
            if isinstance(current_data, np.ndarray) or isinstance(
                current_data, cp.ndarray
            ):
                if self.data_length is None:
                    self.data_length = len(current_data)
                else:
                    assert len(current_data) == self.data_length

                new_out[i] = current_data
            else:
                raise ValueError

        self.nchannels = len(new_out)
        self._data_res_arr = np.asarray(list(new_out), dtype=complex)

    def __getitem__(self, index: tuple) -> np.ndarray:
        return self.data_res_arr[index]

    @property
    def ndim(self) -> int:
        return self.data_res_arr.ndim

    def flatten(self) -> np.ndarray:
        return self.data_res_arr.flatten()

    @property
    def shape(self) -> tuple:
        return self.data_res_arr.shape

    def loglog(
        self,
        ax: Optional[List[plt.Axes] | plt.Axes] = None,
        fig: Optional[plt.Figure] = None,
        inds: Optional[List[int] | int] = None,
        char_strain: Optional[bool] = False,
        **kwargs: dict,
    ) -> Tuple[plt.Figure, plt.Axes]:
        """Produce a log-log plot of the data.

        Args:
            ax: Matplotlib Axes objects to add plots. Either a list of Axes objects or a single Axes object.
            fig: Matplotlib figure object.
            inds: Integer index to select out which data to add to a single access.
                A list can be provided if ax is a list. They must be the same length.
            char_strain: If ``True`` return plot in characteristic strain representation.
            **kwargs: Keyword arguments to be passed to ``loglog`` function in matplotlib.

        Returns:
            Matplotlib figure and axes objects in a 2-tuple.


        """
        if ax is None and fig is None:
            nrows = 1
            ncols = self.shape[0]

            fig, ax = plt.subplots(nrows, ncols, sharex=True, sharey=True)
            ax = ax.ravel()
            inds_list = range(len(ax))

        elif ax is not None:
            if isinstance(ax, list):
                assert len(ax) == np.prod(self.shape[:-1])
                if inds is None:
                    inds_list = list(np.arange(np.prod(self.shape[:-1])))
                else:
                    assert isinstance(inds, list) and len(inds) == len(ax)
                    inds_list = inds

            elif isinstance(ax, plt.Axes):
                assert inds is not None and (
                    isinstance(inds, tuple) or isinstance(inds, int)
                )
                ax = [ax]
                inds_list = [inds]

        elif fig is not None:
            raise NotImplementedError

        for i, ax_tmp in zip(inds_list, ax):
            plot_in = np.abs(self.data_res_arr[i])
            if char_strain:
                plot_in *= self.frequency_arr
            ax_tmp.loglog(self.frequency_arr, plot_in, **kwargs)

        return (fig, ax)

    @property
    def char_strain(self) -> np.ndarray:
        """Characteristic strain representation of the data."""
        return np.sqrt(self.f_arr) * np.abs(self.data_res_arr)
